"""stats-refactor

Revision ID: 4cc87e7fbfdf
Revises: e18ed5fcfedf
Create Date: 2019-05-30 18:40:17.669756

"""

import math
from datetime import timedelta
from decimal import Decimal
from typing import Any, cast

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql.expression import bindparam

from ai.backend.manager.models.base import IDColumn, ResourceSlotColumn, convention

# revision identifiers, used by Alembic.
revision = "4cc87e7fbfdf"
down_revision = "e18ed5fcfedf"
branch_labels = None
depends_on = None


def upgrade() -> None:
    metadata = sa.MetaData(naming_convention=convention)

    # previous table def used for migration
    kernels = sa.Table(
        "kernels",
        metadata,
        # preserved and referred columns
        IDColumn(),
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.func.now(), index=True
        ),
        sa.Column(
            "terminated_at",
            sa.DateTime(timezone=True),
            nullable=True,
            default=sa.null(),
            index=True,
        ),
        sa.Column("occupied_slots", ResourceSlotColumn(), nullable=False),
        sa.Column("occupied_shares", postgresql.JSONB(), nullable=False, default={}),
        # old column(s) to be migrated and removed
        sa.Column("cpu_used", sa.BigInteger(), default=0),  # msec
        sa.Column("mem_max_bytes", sa.BigInteger(), default=0),
        sa.Column("net_rx_bytes", sa.BigInteger(), default=0),
        sa.Column("net_tx_bytes", sa.BigInteger(), default=0),
        sa.Column("io_read_bytes", sa.BigInteger(), default=0),
        sa.Column("io_write_bytes", sa.BigInteger(), default=0),
        sa.Column("io_max_scratch_size", sa.BigInteger(), default=0),
        # new column(s) to be added
        sa.Column("last_stat", postgresql.JSONB(), nullable=True, default=sa.null()),
    )

    op.add_column(
        "kernels", sa.Column("last_stat", postgresql.JSONB(astext_type=sa.Text()), nullable=True)
    )

    connection = op.get_bind()
    query = (
        sa.select(
            kernels.c.id,
            kernels.c.created_at,
            kernels.c.terminated_at,
            kernels.c.occupied_slots,
            kernels.c.occupied_shares,
            kernels.c.cpu_used,
            kernels.c.mem_max_bytes,
            kernels.c.net_rx_bytes,
            kernels.c.net_tx_bytes,
            kernels.c.io_read_bytes,
            kernels.c.io_write_bytes,
            kernels.c.io_max_scratch_size,
        )
        .select_from(kernels)
        .order_by(kernels.c.created_at)
    )
    results = connection.execute(query).fetchall()

    updates = []
    q_pct = Decimal("0.00")
    for row in results:
        row_dict = cast(dict[str, Any], row._mapping)
        if row_dict["terminated_at"] is None:
            cpu_avg_pct: Decimal | int = 0
        else:
            cpu_avg_pct = (
                (
                    Decimal(100)
                    * Decimal(
                        timedelta(microseconds=1e3 * row_dict["cpu_used"])
                        / (row_dict["terminated_at"] - row_dict["created_at"])
                    )
                )
                .quantize(q_pct)
                .normalize()
            )
        mem_capacity = 0
        _oslots = row_dict["occupied_slots"]
        if _oslots:
            mem_capacity = _oslots.get("mem")
        if mem_capacity is None or mem_capacity == 0:
            # fallback: try legacy field
            _oshares = row_dict["occupied_shares"]
            mem_capacity = _oshares.get("mem")
        if mem_capacity is None or mem_capacity == 0:
            # fallback: round-up to nearest GiB
            mem_capacity = math.ceil(row_dict["mem_max_bytes"] / (2**30)) * (2**30)
        if mem_capacity is None or mem_capacity == 0:
            # fallback: assume 1 GiB
            mem_capacity = 2**30
        last_stat = {
            "cpu_used": {
                "current": str(row_dict["cpu_used"]),
                "capacity": None,
            },
            "cpu_util": {
                "current": str(cpu_avg_pct),
                "capacity": None,
                "stats.avg": str(cpu_avg_pct),
            },
            "mem": {
                "current": str(row_dict["mem_max_bytes"]),
                "capacity": str(mem_capacity),
                "stats.max": str(row_dict["mem_max_bytes"]),
            },
            "io_read": {
                "current": str(row_dict["io_read_bytes"]),
                "capacity": None,
                "stats.rate": "0",
            },
            "io_write": {
                "current": str(row_dict["io_write_bytes"]),
                "capacity": None,
                "stats.rate": "0",
            },
            "io_scratch_size": {
                "current": str(row_dict["io_max_scratch_size"]),
                "capacity": str(10 * (2**30)),  # 10 GiB
                "stats.max": str(row_dict["io_max_scratch_size"]),
            },
        }
        updates.append({"row_id": row_dict["id"], "last_stat": last_stat})

    if updates:
        update_query = (
            sa.update(kernels)
            .values(last_stat=bindparam("last_stat"))
            .where(kernels.c.id == bindparam("row_id"))
        )
        connection.execute(update_query, updates)

    op.drop_column("kernels", "io_max_scratch_size")
    op.drop_column("kernels", "net_rx_bytes")
    op.drop_column("kernels", "net_tx_bytes")
    op.drop_column("kernels", "mem_max_bytes")
    op.drop_column("kernels", "io_write_bytes")
    op.drop_column("kernels", "io_read_bytes")
    op.drop_column("kernels", "cpu_used")


def downgrade() -> None:
    metadata = sa.MetaData(naming_convention=convention)

    kernels = sa.Table(
        "kernels",
        metadata,
        # preserved and referred columns
        IDColumn(),
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.func.now(), index=True
        ),
        sa.Column(
            "terminated_at",
            sa.DateTime(timezone=True),
            nullable=True,
            default=sa.null(),
            index=True,
        ),
        sa.Column("occupied_slots", ResourceSlotColumn(), nullable=False),
        sa.Column("occupied_shares", postgresql.JSONB(), nullable=False, default={}),
        # old column(s) to be migrated
        sa.Column("last_stat", postgresql.JSONB(), nullable=True, default=sa.null()),
        # new column(s) to be added
        sa.Column("cpu_used", sa.BigInteger(), default=0),  # msec
        sa.Column("mem_max_bytes", sa.BigInteger(), default=0),
        sa.Column("net_rx_bytes", sa.BigInteger(), default=0),
        sa.Column("net_tx_bytes", sa.BigInteger(), default=0),
        sa.Column("io_read_bytes", sa.BigInteger(), default=0),
        sa.Column("io_write_bytes", sa.BigInteger(), default=0),
        sa.Column("io_max_scratch_size", sa.BigInteger(), default=0),
    )

    op.add_column("kernels", sa.Column("cpu_used", sa.BIGINT(), autoincrement=False, nullable=True))
    op.add_column(
        "kernels", sa.Column("io_read_bytes", sa.BIGINT(), autoincrement=False, nullable=True)
    )
    op.add_column(
        "kernels", sa.Column("io_write_bytes", sa.BIGINT(), autoincrement=False, nullable=True)
    )
    op.add_column(
        "kernels", sa.Column("mem_max_bytes", sa.BIGINT(), autoincrement=False, nullable=True)
    )
    op.add_column(
        "kernels", sa.Column("net_tx_bytes", sa.BIGINT(), autoincrement=False, nullable=True)
    )
    op.add_column(
        "kernels", sa.Column("net_rx_bytes", sa.BIGINT(), autoincrement=False, nullable=True)
    )
    op.add_column(
        "kernels", sa.Column("io_max_scratch_size", sa.BIGINT(), autoincrement=False, nullable=True)
    )

    # Restore old stats
    connection = op.get_bind()
    query = sa.select(kernels.c.id, kernels.c.last_stat).select_from(kernels)
    results = connection.execute(query).fetchall()
    updates = []
    for row in results:
        row_dict = cast(dict[str, Any], row._mapping)
        last_stat = row_dict["last_stat"]
        updates.append({
            "row_id": row_dict["id"],
            "cpu_used": Decimal(last_stat["cpu_used"]["current"]),
            "io_read_bytes": int(last_stat["io_read"]["current"]),
            "io_write_bytes": int(last_stat["io_write"]["current"]),
            "mem_max_bytes": int(last_stat["mem"]["stats.max"]),
            "io_max_scratch_size": int(last_stat["io_scratch_size"]["stats.max"]),
        })
    if updates:
        update_query = (
            sa.update(kernels)
            .values({
                "cpu_used": bindparam("cpu_used"),
                "io_read_bytes": bindparam("io_read_bytes"),
                "io_write_bytes": bindparam("io_write_bytes"),
                "mem_max_bytes": bindparam("mem_max_bytes"),
                "net_tx_bytes": 0,
                "net_rx_bytes": 0,
                "io_max_scratch_size": bindparam("io_max_scratch_size"),
            })
            .where(kernels.c.id == bindparam("row_id"))
        )
        connection.execute(update_query, updates)

    op.drop_column("kernels", "last_stat")



"""Deployment resolver package.

Contains GraphQL query, mutation, and subscription resolver functions
for the deployment module.
"""

from .access_token import (
    create_access_token,
)
from .auto_scaling import (
    create_auto_scaling_rule,
    delete_auto_scaling_rule,
    update_auto_scaling_rule,
)
from .deployment import (
    create_model_deployment,
    delete_model_deployment,
    deployment,
    deployment_status_changed,
    deployments,
    sync_replicas,
    update_model_deployment,
)
from .replica import (
    replica,
    replica_status_changed,
    replicas,
)
from .revision import (
    activate_deployment_revision,
    add_model_revision,
    create_model_revision,
    inference_runtime_config,
    inference_runtime_configs,
    revision,
    revisions,
)
from .route import (
    route,
    routes,
    update_route_traffic_status,
)

__all__ = [
    # Access Token
    "create_access_token",
    # Auto Scaling
    "create_auto_scaling_rule",
    "update_auto_scaling_rule",
    "delete_auto_scaling_rule",
    # Deployment
    "deployments",
    "deployment",
    "create_model_deployment",
    "update_model_deployment",
    "delete_model_deployment",
    "sync_replicas",
    "deployment_status_changed",
    # Replica
    "replicas",
    "replica",
    "replica_status_changed",
    # Revision
    "revisions",
    "revision",
    "inference_runtime_config",
    "inference_runtime_configs",
    "add_model_revision",
    "create_model_revision",
    "activate_deployment_revision",
    # Route
    "routes",
    "route",
    "update_route_traffic_status",
]



"""add a new session type INFERENCE

Revision ID: b5be363ab05c
Revises: b6b884fbae1f
Create Date: 2023-02-15 15:24:29.112938

Reference: https://www.postgresql.org/message-id/CANu8FiwwBxZZGX23=Na_7bc4DZ-yzd_poKhaoPmN3+SHG08MAg@mail.gmail.com

"""

import textwrap

from alembic import op
from sqlalchemy.sql import text

from ai.backend.common.types import SessionTypes

# revision identifiers, used by Alembic.
revision = "b5be363ab05c"
down_revision = "b6b884fbae1f"
branch_labels = None
depends_on = None

# select enum_range(null::sessiontypes);
typename = SessionTypes.__name__.lower()


def upgrade() -> None:
    op.execute(text(f"ALTER TYPE {typename} ADD VALUE IF NOT EXISTS 'INFERENCE';"))
    op.execute(
        text(
            textwrap.dedent(
                f"""\
                ALTER TABLE sessions ALTER COLUMN session_type TYPE {typename} USING session_type::text::sessiontypes;
            """
            )
        )
    )


def downgrade() -> None:
    conn = op.get_bind()
    conn.execute(
        text(
            f"UPDATE sessions SET session_type = 'INTERACTIVE'::{typename} WHERE session_type ="
            f" 'INFERENCE'::{typename};"
        )
    )

    cursor = conn.execute(
        text(
            textwrap.dedent(
                """\
                SELECT t.typname, e.enumlabel, e.enumsortorder, e.enumtypid
                FROM pg_type t
                JOIN pg_enum e ON e.enumtypid = t.oid
                WHERE t.typtype = 'e'
                AND e.enumlabel = 'INFERENCE'
                ORDER BY 1, enumsortorder;
            """
            )
        )
    )

    if (row := next(cursor, None)) is None:
        return

    typname, enumlabel, enumsortorder, enumtypid = row

    conn.execute(
        text(
            textwrap.dedent(
                f"""\
                DELETE FROM pg_enum
                WHERE enumtypid = {enumtypid}
                AND enumlabel = 'INFERENCE';
            """
            )
        )
    )



from dataclasses import dataclass
from typing import Any, override

from aiohttp import MultipartReader

from ai.backend.common.types import AccessKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.session.types import SessionData
from ai.backend.manager.services.session.actions.file_base import SessionFileAction


@dataclass
class UploadFilesAction(SessionFileAction):
    session_name: str
    owner_access_key: AccessKey
    # TODO: Refactor this.
    reader: MultipartReader

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class UploadFilesActionResult(BaseActionResult):
    # TODO: Add proper type
    result: Any
    session_data: SessionData

    @override
    def entity_id(self) -> str | None:
        return str(self.session_data.id)



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.types import RepositoryArgs

from .repository import StorageNamespaceRepository


@dataclass
class StorageNamespaceRepositories:
    repository: StorageNamespaceRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        return cls(
            repository=StorageNamespaceRepository(
                db=args.db,
            ),
        )



"""CreatorSpec implementations for image repository."""

from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID

from ai.backend.manager.models.image.row import ImageAliasRow
from ai.backend.manager.repositories.base import CreatorSpec


@dataclass
class ImageAliasCreatorSpec(CreatorSpec[ImageAliasRow]):
    """CreatorSpec for image alias creation."""

    alias: str
    image_id: UUID

    def build_row(self) -> ImageAliasRow:
        return ImageAliasRow(
            alias=self.alias,
            image_id=self.image_id,
        )



"""Query conditions and orders for user repository operations."""

from __future__ import annotations

from collections.abc import Collection
from datetime import datetime
from typing import Any

import sqlalchemy as sa

from ai.backend.common.data.user.types import UserRole
from ai.backend.manager.api.gql.base import StringMatchSpec, UUIDEqualMatchSpec, UUIDInMatchSpec
from ai.backend.manager.data.user.types import UserStatus
from ai.backend.manager.models.domain import DomainRow
from ai.backend.manager.models.group import AssocGroupUserRow, GroupRow
from ai.backend.manager.models.user import UserRow
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder

__all__ = (
    "UserConditions",
    "UserOrders",
)


class UserConditions:
    """Query conditions for filtering users."""

    # ==================== Email Filters ====================

    @staticmethod
    def by_email_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = UserRow.email.ilike(f"%{spec.value}%")
            else:
                condition = UserRow.email.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_email_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(UserRow.email) == spec.value.lower()
            else:
                condition = UserRow.email == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_email_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = UserRow.email.ilike(f"{spec.value}%")
            else:
                condition = UserRow.email.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_email_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = UserRow.email.ilike(f"%{spec.value}")
            else:
                condition = UserRow.email.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    # ==================== Username Filters ====================

    @staticmethod
    def by_username_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = UserRow.username.ilike(f"%{spec.value}%")
            else:
                condition = UserRow.username.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_username_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(UserRow.username) == spec.value.lower()
            else:
                condition = UserRow.username == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_username_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = UserRow.username.ilike(f"{spec.value}%")
            else:
                condition = UserRow.username.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_username_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = UserRow.username.ilike(f"%{spec.value}")
            else:
                condition = UserRow.username.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    # ==================== Domain Name Filters ====================

    @staticmethod
    def by_domain_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = UserRow.domain_name.ilike(f"%{spec.value}%")
            else:
                condition = UserRow.domain_name.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_domain_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(UserRow.domain_name) == spec.value.lower()
            else:
                condition = UserRow.domain_name == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_domain_name_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = UserRow.domain_name.ilike(f"{spec.value}%")
            else:
                condition = UserRow.domain_name.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_domain_name_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = UserRow.domain_name.ilike(f"%{spec.value}")
            else:
                condition = UserRow.domain_name.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    # ==================== Status Filters ====================

    @staticmethod
    def by_is_active(is_active: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if is_active:
                return UserRow.status == UserStatus.ACTIVE
            return UserRow.status != UserStatus.ACTIVE

        return inner

    @staticmethod
    def by_status_equals(status: UserStatus) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.status == status

        return inner

    @staticmethod
    def by_status_in(statuses: Collection[UserStatus]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.status.in_(statuses)

        return inner

    # ==================== Role Filters ====================

    @staticmethod
    def by_role_equals(role: UserRole) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.role == role

        return inner

    @staticmethod
    def by_role_in(roles: Collection[UserRole]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.role.in_(roles)

        return inner

    # ==================== UUID Filters ====================

    @staticmethod
    def by_uuid_equals(spec: UUIDEqualMatchSpec) -> QueryCondition:
        """Filter by UUID equality.

        Args:
            spec: UUID equality specification with value and negated flag.

        Returns:
            QueryCondition callable for UUID filtering.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            condition = UserRow.uuid == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_uuid_in(spec: UUIDInMatchSpec) -> QueryCondition:
        """Filter by UUID IN operation.

        Args:
            spec: UUID IN specification with values list and negated flag.

        Returns:
            QueryCondition callable for UUID IN filtering.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            condition = UserRow.uuid.in_(spec.values)
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    # ==================== DateTime Filters ====================

    @staticmethod
    def by_created_at_before(dt: datetime) -> QueryCondition:
        """Filter by created_at < datetime."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.created_at < dt

        return inner

    @staticmethod
    def by_created_at_after(dt: datetime) -> QueryCondition:
        """Filter by created_at > datetime."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.created_at > dt

        return inner

    @staticmethod
    def by_modified_at_before(dt: datetime) -> QueryCondition:
        """Filter by modified_at < datetime."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.modified_at < dt

        return inner

    @staticmethod
    def by_modified_at_after(dt: datetime) -> QueryCondition:
        """Filter by modified_at > datetime."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.modified_at > dt

        return inner

    # ==================== Cursor Pagination ====================

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(UserRow.created_at).where(UserRow.uuid == cursor_id).scalar_subquery()
            )
            return UserRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(UserRow.created_at).where(UserRow.uuid == cursor_id).scalar_subquery()
            )
            return UserRow.created_at > subquery

        return inner

    # ==================== Domain Nested Filters ====================

    @staticmethod
    def _exists_domain(
        *domain_conditions: sa.sql.expression.ColumnElement[bool],
    ) -> sa.sql.expression.ColumnElement[bool]:
        """EXISTS subquery: User → Domain (via FK domain_name)."""
        subq = sa.select(sa.literal(1)).where(DomainRow.name == UserRow.domain_name)
        for cond in domain_conditions:
            subq = subq.where(cond)
        return sa.exists(subq)

    @staticmethod
    def by_domain_description_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                cond = DomainRow.description.ilike(f"%{spec.value}%")
            else:
                cond = DomainRow.description.like(f"%{spec.value}%")
            if spec.negated:
                cond = sa.not_(cond)
            return UserConditions._exists_domain(cond)

        return inner

    @staticmethod
    def by_domain_description_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                cond = sa.func.lower(DomainRow.description) == spec.value.lower()
            else:
                cond = DomainRow.description == spec.value
            if spec.negated:
                cond = sa.not_(cond)
            return UserConditions._exists_domain(cond)

        return inner

    @staticmethod
    def by_domain_is_active(is_active: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserConditions._exists_domain(DomainRow.is_active == is_active)

        return inner

    @staticmethod
    def exists_domain_combined(domain_conditions: list[QueryCondition]) -> QueryCondition:
        """Combine multiple domain conditions into single EXISTS subquery.

        Accepts conditions from DomainV2Filter.build_conditions() directly.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subq = sa.select(sa.literal(1)).where(DomainRow.name == UserRow.domain_name)
            for cond in domain_conditions:
                subq = subq.where(cond())
            return sa.exists(subq)

        return inner

    # ==================== Project Nested Filters (M:N) ====================

    @staticmethod
    def _exists_project(
        *project_conditions: sa.sql.expression.ColumnElement[bool],
    ) -> sa.sql.expression.ColumnElement[bool]:
        """EXISTS subquery: User → Project (via M:N AssocGroupUserRow)."""
        subq = (
            sa.select(sa.literal(1))
            .select_from(
                sa.join(
                    AssocGroupUserRow.__table__,
                    GroupRow.__table__,
                    AssocGroupUserRow.group_id == GroupRow.id,
                )
            )
            .where(AssocGroupUserRow.user_id == UserRow.uuid)
        )
        for cond in project_conditions:
            subq = subq.where(cond)
        return sa.exists(subq)

    @staticmethod
    def by_project_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                cond = GroupRow.name.ilike(f"%{spec.value}%")
            else:
                cond = GroupRow.name.like(f"%{spec.value}%")
            if spec.negated:
                cond = sa.not_(cond)
            return UserConditions._exists_project(cond)

        return inner

    @staticmethod
    def by_project_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                cond = sa.func.lower(GroupRow.name) == spec.value.lower()
            else:
                cond = GroupRow.name == spec.value
            if spec.negated:
                cond = sa.not_(cond)
            return UserConditions._exists_project(cond)

        return inner

    @staticmethod
    def by_project_is_active(is_active: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserConditions._exists_project(GroupRow.is_active == is_active)

        return inner

    @staticmethod
    def exists_project_combined(project_conditions: list[QueryCondition]) -> QueryCondition:
        """Combine multiple project conditions into single EXISTS subquery.

        Accepts conditions from ProjectV2Filter.build_conditions() directly.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subq = (
                sa.select(sa.literal(1))
                .select_from(
                    sa.join(
                        AssocGroupUserRow.__table__,
                        GroupRow.__table__,
                        AssocGroupUserRow.group_id == GroupRow.id,
                    )
                )
                .where(AssocGroupUserRow.user_id == UserRow.uuid)
            )
            for cond in project_conditions:
                subq = subq.where(cond())
            return sa.exists(subq)

        return inner


class UserOrders:
    """Query orders for sorting users."""

    @staticmethod
    def uuid(ascending: bool = True) -> QueryOrder:
        if ascending:
            return UserRow.uuid.asc()
        return UserRow.uuid.desc()

    @staticmethod
    def email(ascending: bool = True) -> QueryOrder:
        if ascending:
            return UserRow.email.asc()
        return UserRow.email.desc()

    @staticmethod
    def username(ascending: bool = True) -> QueryOrder:
        if ascending:
            return UserRow.username.asc()
        return UserRow.username.desc()

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return UserRow.created_at.asc()
        return UserRow.created_at.desc()

    @staticmethod
    def modified_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return UserRow.modified_at.asc()
        return UserRow.modified_at.desc()

    @staticmethod
    def status(ascending: bool = True) -> QueryOrder:
        if ascending:
            return UserRow.status.asc()
        return UserRow.status.desc()

    @staticmethod
    def domain_name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return UserRow.domain_name.asc()
        return UserRow.domain_name.desc()

    # ==================== Domain Nested Orders ====================

    @staticmethod
    def _scalar_domain(
        column: sa.ColumnElement[Any] | sa.orm.InstrumentedAttribute[Any],
    ) -> sa.ScalarSelect[Any]:
        """Scalar subquery selecting a Domain column correlated to current User."""
        return (
            sa.select(column)
            .where(DomainRow.name == UserRow.domain_name)
            .correlate(UserRow)
            .scalar_subquery()
        )

    @staticmethod
    def by_domain_name(ascending: bool = True) -> QueryOrder:
        subq = UserOrders._scalar_domain(DomainRow.name)
        return subq.asc() if ascending else subq.desc()

    @staticmethod
    def by_domain_created_at(ascending: bool = True) -> QueryOrder:
        subq = UserOrders._scalar_domain(DomainRow.created_at)
        return subq.asc() if ascending else subq.desc()

    # ==================== Project Nested Orders (M:N → MIN aggregation) ====================

    @staticmethod
    def _scalar_project_min(
        column: sa.ColumnElement[Any] | sa.orm.InstrumentedAttribute[Any],
    ) -> sa.ScalarSelect[Any]:
        """Scalar subquery with MIN for M:N relationship."""
        return (
            sa.select(sa.func.min(column))
            .select_from(
                sa.join(
                    AssocGroupUserRow.__table__,
                    GroupRow.__table__,
                    AssocGroupUserRow.group_id == GroupRow.id,
                )
            )
            .where(AssocGroupUserRow.user_id == UserRow.uuid)
            .correlate(UserRow)
            .scalar_subquery()
        )

    @staticmethod
    def by_project_name(ascending: bool = True) -> QueryOrder:
        subq = UserOrders._scalar_project_min(GroupRow.name)
        return subq.asc() if ascending else subq.desc()



"""Resource Usage History Repositories wrapper."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.types import RepositoryArgs

from .repository import ResourceUsageHistoryRepository

__all__ = ("ResourceUsageHistoryRepositories",)


@dataclass
class ResourceUsageHistoryRepositories:
    repository: ResourceUsageHistoryRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = ResourceUsageHistoryRepository(args.db)

        return cls(
            repository=repository,
        )



"""CreatorSpec implementations for keypair resource policy repository."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, override

from ai.backend.common.types import DefaultForUnspecified, ResourceSlot
from ai.backend.manager.models.resource_policy import KeyPairResourcePolicyRow
from ai.backend.manager.repositories.base import CreatorSpec


@dataclass
class KeyPairResourcePolicyCreatorSpec(CreatorSpec["KeyPairResourcePolicyRow"]):
    """CreatorSpec for keypair resource policy creation."""

    name: str | None
    allowed_vfolder_hosts: dict[str, Any] | None
    default_for_unspecified: DefaultForUnspecified | None
    idle_timeout: int | None
    max_concurrent_sessions: int | None
    max_containers_per_session: int | None
    max_pending_session_count: int | None
    max_pending_session_resource_slots: ResourceSlot | None
    max_quota_scope_size: int | None
    max_vfolder_count: int | None
    max_vfolder_size: int | None
    max_concurrent_sftp_sessions: int | None
    max_session_lifetime: int | None
    total_resource_slots: ResourceSlot | None

    @override
    def build_row(self) -> KeyPairResourcePolicyRow:
        return KeyPairResourcePolicyRow(
            name=self.name,
            default_for_unspecified=self.default_for_unspecified,
            total_resource_slots=self.total_resource_slots,
            max_session_lifetime=self.max_session_lifetime,
            max_concurrent_sessions=self.max_concurrent_sessions,
            max_pending_session_count=self.max_pending_session_count,
            max_pending_session_resource_slots=self.max_pending_session_resource_slots,
            max_concurrent_sftp_sessions=self.max_concurrent_sftp_sessions,
            max_containers_per_session=self.max_containers_per_session,
            idle_timeout=self.idle_timeout,
            allowed_vfolder_hosts=self.allowed_vfolder_hosts,
        )



"""Fair share calculation module for the Sokovan scheduler.

This module provides components for tracking kernel resource usage
and calculating fair share scheduling ranks.
"""

from .aggregator import (
    DomainUsageBucketKey,
    FairShareAggregator,
    KernelUsagePreparationResult,
    ProjectUsageBucketKey,
    UsageBucketAggregationResult,
    UserUsageBucketKey,
)
from .calculator import (
    DomainFactorResult,
    FairShareFactorCalculationResult,
    FairShareFactorCalculator,
    ProjectFactorResult,
    UserFactorResult,
    UserSchedulingRank,
)

__all__ = [
    # Aggregator
    "DomainUsageBucketKey",
    "FairShareAggregator",
    "KernelUsagePreparationResult",
    "ProjectUsageBucketKey",
    "UsageBucketAggregationResult",
    "UserUsageBucketKey",
    # Calculator
    "DomainFactorResult",
    "FairShareFactorCalculationResult",
    "FairShareFactorCalculator",
    "ProjectFactorResult",
    "UserFactorResult",
    "UserSchedulingRank",
]



import enum
from collections.abc import Mapping
from typing import Any, NamedTuple

import sqlalchemy as sa
from lark import Lark, LarkError, Transformer
from lark.lexer import Token

from . import JSONFieldItem, OrderSpecItem, get_col_from_table

__all__ = (
    "ColumnMapType",
    "QueryOrderParser",
)

_grammar = r"""
    ?start: expr
    expr   : [col ("," col)*]
    col    : ORDER? CNAME
    ORDER  : "+" | "-"
    %import common.CNAME
    %import common.WS
    %ignore WS
"""
_parser = Lark(
    _grammar,
    parser="lalr",
    maybe_placeholders=False,
)

type ColumnMapType = Mapping[str, OrderSpecItem] | None


class OrderDirection(enum.Enum):
    ASC = "asc"
    DESC = "desc"


class OrderingItem(NamedTuple):
    column: (
        sa.Column[Any]
        | sa.orm.attributes.InstrumentedAttribute[Any]
        | sa.sql.elements.KeyedColumnElement[Any]
    )
    order_direction: OrderDirection


class QueryOrderTransformer(Transformer[Any, Any]):
    def __init__(
        self, sa_table: sa.Table | sa.sql.Join | type, column_map: ColumnMapType | None = None
    ) -> None:
        super().__init__()
        self._sa_table = sa_table
        self._column_map = column_map

    def _get_col(
        self, col_name: str
    ) -> (
        sa.Column[Any]
        | sa.orm.attributes.InstrumentedAttribute[Any]
        | sa.sql.elements.KeyedColumnElement[Any]
    ):
        try:
            if self._column_map:
                col_value, func = self._column_map[col_name]
                match col_value:
                    case str(column):
                        matched_col = get_col_from_table(self._sa_table, column)
                    case JSONFieldItem(_col, _key):
                        _column = get_col_from_table(self._sa_table, _col)
                        matched_col = _column.op("->>")(_key)  # type: ignore[assignment]
                    case _:
                        raise ValueError("Invalid type of field name", col_name)
                col = func(matched_col) if func is not None else matched_col  # type: ignore[arg-type]
            else:
                col = get_col_from_table(self._sa_table, col_name)
            return col
        except KeyError as e:
            raise ValueError("Unknown/unsupported field name", col_name) from e

    def col(self, *args: Any) -> OrderingItem:
        children: list[Token] = args[0]
        if len(children) == 2:
            op = children[0].value
            col = self._get_col(children[1].value)
        else:
            op = "+"  # assume ascending if not marked
            col = self._get_col(children[0].value)
        if op == "+":
            return OrderingItem(col, OrderDirection.ASC)
        if op == "-":
            return OrderingItem(col, OrderDirection.DESC)
        raise ValueError(f"Invalid operation `{op}`. Please use `+` or `-`")

    expr = tuple


class QueryOrderParser:
    def __init__(self, column_map: ColumnMapType | None = None) -> None:
        self._column_map = column_map
        self._parser = _parser

    def parse_order(
        self, table: sa.Table | sa.sql.Join | type, order_expr: str
    ) -> list[OrderingItem]:
        try:
            ast = self._parser.parse(order_expr)
            result: list[OrderingItem] = QueryOrderTransformer(table, self._column_map).transform(
                ast
            )
            return result
        except LarkError as e:
            raise ValueError(f"Query ordering parsing error: {e}") from e

    def append_ordering(
        self,
        sa_query: sa.sql.Select[Any],
        order_expr: str,
    ) -> sa.sql.Select[Any]:
        """
        Parse the given filter expression and build the where clause based on the first target table from
        the given SQLAlchemy query object.
        """
        table = sa_query.froms[0]
        # FromClause is compatible with our union type, cast for type checker
        from typing import cast

        parsed_table = cast(sa.Table | sa.sql.Join | type, table)
        orders = [
            col.asc() if direction == OrderDirection.ASC else col.desc()
            for col, direction in self.parse_order(parsed_table, order_expr)
        ]
        return sa_query.order_by(*orders)



"""GraphQL resource group module."""

from .resolver import (
    admin_resource_groups,
    admin_update_resource_group,
    admin_update_resource_group_fair_share_spec,
    resource_groups,
    update_resource_group_fair_share_spec,
)
from .types import (
    FairShareScalingGroupSpecGQL,
    ResourceGroupFilterGQL,
    ResourceGroupGQL,
    ResourceGroupOrderByGQL,
    ResourceGroupOrderFieldGQL,
    UpdateResourceGroupFairShareSpecInput,
    UpdateResourceGroupFairShareSpecPayload,
    UpdateResourceGroupInput,
    UpdateResourceGroupPayload,
)

__all__ = (
    # Types
    "FairShareScalingGroupSpecGQL",
    "ResourceGroupGQL",
    "ResourceGroupFilterGQL",
    "ResourceGroupOrderByGQL",
    "ResourceGroupOrderFieldGQL",
    "UpdateResourceGroupFairShareSpecInput",
    "UpdateResourceGroupFairShareSpecPayload",
    "UpdateResourceGroupInput",
    "UpdateResourceGroupPayload",
    # Query Resolvers - Admin
    "admin_resource_groups",
    # Query Resolvers - Legacy (deprecated)
    "resource_groups",
    # Mutation Resolvers - Admin
    "admin_update_resource_group_fair_share_spec",
    "admin_update_resource_group",
    # Mutation Resolvers - Legacy (deprecated)
    "update_resource_group_fair_share_spec",
)



import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.association.types import AssociationArtifactsStoragesData
from ai.backend.manager.services.artifact_revision.actions.base import ArtifactRevisionAction


@dataclass
class DisassociateWithStorageAction(ArtifactRevisionAction):
    artifact_revision_id: uuid.UUID
    storage_namespace_id: uuid.UUID

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.ARTIFACT_REVISION_STORAGE_LINK

    @override
    def entity_id(self) -> str | None:
        return str(self.artifact_revision_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE


@dataclass
class DisassociateWithStorageActionResult(BaseActionResult):
    result: AssociationArtifactsStoragesData

    @override
    def entity_id(self) -> str | None:
        return str(self.result.id)



"""Handler for provisioning routes."""

import logging
from collections.abc import Sequence

from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.deployment.types import RouteStatus, RouteStatusTransitions
from ai.backend.manager.defs import LockID
from ai.backend.manager.repositories.deployment.types import RouteData
from ai.backend.manager.sokovan.deployment.route.executor import RouteExecutor
from ai.backend.manager.sokovan.deployment.route.types import RouteExecutionResult

from .base import RouteHandler

log = BraceStyleAdapter(logging.getLogger(__name__))


class ProvisioningRouteHandler(RouteHandler):
    """Handler for provisioning routes (PROVISIONING -> UNHEALTHY)."""

    def __init__(
        self,
        route_executor: RouteExecutor,
        event_producer: EventProducer,
    ) -> None:
        self._route_executor = route_executor
        self._event_producer = event_producer

    @classmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        return "provision-routes"

    @property
    def lock_id(self) -> LockID | None:
        """Lock for provisioning routes."""
        return LockID.LOCKID_DEPLOYMENT_PROVISIONING_ROUTES

    @classmethod
    def target_statuses(cls) -> list[RouteStatus]:
        """Get the target route statuses for this handler."""
        return [RouteStatus.PROVISIONING]

    @classmethod
    def next_status(cls) -> RouteStatus | None:
        """Get the next route status after this handler's operation."""
        return RouteStatus.DEGRADED

    @classmethod
    def failure_status(cls) -> RouteStatus | None:
        """Get the failure route status if applicable."""
        return RouteStatus.FAILED_TO_START

    @classmethod
    def stale_status(cls) -> RouteStatus | None:
        """Get the stale route status if applicable."""
        return None

    @classmethod
    def status_transitions(cls) -> RouteStatusTransitions:
        """Define state transitions for provisioning route handler (BEP-1030).

        - success: Route → DEGRADED (provisioned, needs health check)
        - failure: Route → FAILED_TO_START
        - stale: None
        """
        return RouteStatusTransitions(
            success=RouteStatus.DEGRADED,
            failure=RouteStatus.FAILED_TO_START,
            stale=None,
        )

    async def execute(self, routes: Sequence[RouteData]) -> RouteExecutionResult:
        """Execute provisioning for routes."""
        log.debug("Provisioning {} routes", len(routes))

        # Execute route provisioning logic via executor
        return await self._route_executor.provision_routes(routes)

    async def post_process(self, result: RouteExecutionResult) -> None:
        """Handle post-processing after provisioning routes."""
        log.info(
            "Provisioned {} routes successfully, {} failed",
            len(result.successes),
            len(result.errors),
        )



"""GraphQL artifact module."""

from .resolver import (
    approve_artifact_revision,
    artifact,
    artifact_import_progress_updated,
    artifact_revision,
    artifact_revisions,
    artifact_status_changed,
    artifacts,
    cancel_import_artifact,
    cleanup_artifact_revisions,
    delegate_import_artifacts,
    delegate_scan_artifacts,
    delete_artifacts,
    import_artifacts,
    reject_artifact_revision,
    restore_artifacts,
    scan_artifact_models,
    scan_artifacts,
    update_artifact,
)
from .types import (
    ApproveArtifactInput,
    ApproveArtifactPayload,
    Artifact,
    ArtifactConnection,
    ArtifactFilter,
    ArtifactOrderBy,
    ArtifactRevision,
    ArtifactRevisionConnection,
    ArtifactRevisionFilter,
    ArtifactRevisionOrderBy,
    CancelArtifactInput,
    CancelImportArtifactPayload,
    CleanupArtifactRevisionsInput,
    CleanupArtifactRevisionsPayload,
    DelegateImportArtifactsInput,
    DelegateImportArtifactsPayload,
    DelegateScanArtifactsInput,
    DelegateScanArtifactsPayload,
    DeleteArtifactsInput,
    DeleteArtifactsPayload,
    ImportArtifactsInput,
    ImportArtifactsPayload,
    RejectArtifactInput,
    RejectArtifactPayload,
    RestoreArtifactsInput,
    RestoreArtifactsPayload,
    ScanArtifactModelsInput,
    ScanArtifactModelsPayload,
    ScanArtifactsInput,
    ScanArtifactsPayload,
    UpdateArtifactInput,
    UpdateArtifactPayload,
)

__all__ = (
    # Main types
    "Artifact",
    "ArtifactRevision",
    "ArtifactConnection",
    "ArtifactRevisionConnection",
    # Filters and ordering
    "ArtifactFilter",
    "ArtifactOrderBy",
    "ArtifactRevisionFilter",
    "ArtifactRevisionOrderBy",
    # Input types and payloads
    "ScanArtifactsInput",
    "ScanArtifactsPayload",
    "ImportArtifactsInput",
    "ImportArtifactsPayload",
    "DelegateScanArtifactsInput",
    "DelegateScanArtifactsPayload",
    "DelegateImportArtifactsInput",
    "DelegateImportArtifactsPayload",
    "UpdateArtifactInput",
    "UpdateArtifactPayload",
    "CleanupArtifactRevisionsInput",
    "CleanupArtifactRevisionsPayload",
    "DeleteArtifactsInput",
    "DeleteArtifactsPayload",
    "RestoreArtifactsInput",
    "RestoreArtifactsPayload",
    "ApproveArtifactInput",
    "ApproveArtifactPayload",
    "RejectArtifactInput",
    "RejectArtifactPayload",
    "CancelArtifactInput",
    "CancelImportArtifactPayload",
    "ScanArtifactModelsInput",
    "ScanArtifactModelsPayload",
    # Query resolvers
    "artifacts",
    "artifact",
    "artifact_revisions",
    "artifact_revision",
    # Mutation resolvers
    "scan_artifacts",
    "import_artifacts",
    "delegate_scan_artifacts",
    "delegate_import_artifacts",
    "update_artifact",
    "cleanup_artifact_revisions",
    "delete_artifacts",
    "restore_artifacts",
    "cancel_import_artifact",
    "approve_artifact_revision",
    "reject_artifact_revision",
    "scan_artifact_models",
    # Subscription resolvers
    "artifact_status_changed",
    "artifact_import_progress_updated",
)



"""User service actions.

Re-exports all action types from submodules.
"""

from .admin_month_stats import (
    AdminMonthStatsAction,
    AdminMonthStatsActionResult,
)
from .create_user import (
    CreateUserAction,
    CreateUserActionResult,
)
from .delete_user import (
    DeleteUserAction,
    DeleteUserActionResult,
)
from .get_user import (
    GetUserAction,
    GetUserActionResult,
)
from .modify_user import (
    ModifyUserAction,
    ModifyUserActionResult,
)
from .purge_user import (
    PurgeUserAction,
    PurgeUserActionResult,
)
from .search_users import (
    SearchUsersAction,
    SearchUsersActionResult,
)
from .search_users_by_domain import (
    SearchUsersByDomainAction,
    SearchUsersByDomainActionResult,
)
from .search_users_by_project import (
    SearchUsersByProjectAction,
    SearchUsersByProjectActionResult,
)
from .user_month_stats import (
    UserMonthStatsAction,
    UserMonthStatsActionResult,
)

__all__ = (
    "AdminMonthStatsAction",
    "AdminMonthStatsActionResult",
    "CreateUserAction",
    "CreateUserActionResult",
    "DeleteUserAction",
    "DeleteUserActionResult",
    "GetUserAction",
    "GetUserActionResult",
    "ModifyUserAction",
    "ModifyUserActionResult",
    "PurgeUserAction",
    "PurgeUserActionResult",
    "SearchUsersAction",
    "SearchUsersActionResult",
    "SearchUsersByDomainAction",
    "SearchUsersByDomainActionResult",
    "SearchUsersByProjectAction",
    "SearchUsersByProjectActionResult",
    "UserMonthStatsAction",
    "UserMonthStatsActionResult",
)



"""Types for group repository operations."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from uuid import UUID

import sqlalchemy as sa

from ai.backend.manager.data.group.types import GroupData
from ai.backend.manager.errors.resource import DomainNotFound
from ai.backend.manager.models.domain import DomainRow
from ai.backend.manager.models.group.row import AssocGroupUserRow, GroupRow
from ai.backend.manager.repositories.base import ExistenceCheck, QueryCondition, SearchScope

__all__ = (
    "GroupSearchResult",
    "DomainProjectSearchScope",
    "UserProjectSearchScope",
)


@dataclass
class GroupSearchResult:
    """Result from searching groups/projects."""

    items: list[GroupData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


@dataclass(frozen=True)
class DomainProjectSearchScope(SearchScope):
    """Required scope for searching projects within a domain.

    Used for domain-scoped project search (domain admin+).
    """

    domain_name: str
    """Required. The domain to search within."""

    def to_condition(self) -> QueryCondition:
        """Convert scope to a query condition for GroupRow."""
        domain_name = self.domain_name

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return GroupRow.domain_name == domain_name

        return inner

    @property
    def existence_checks(self) -> Sequence[ExistenceCheck[str]]:
        """Return existence checks for scope validation."""
        return [
            ExistenceCheck(
                column=DomainRow.name,
                value=self.domain_name,
                error=DomainNotFound(self.domain_name),
            ),
        ]


@dataclass(frozen=True)
class UserProjectSearchScope(SearchScope):
    """Required scope for searching projects a user is member of.

    Used for user-scoped project search (any authenticated user).
    Requires checking association_groups_users table.
    """

    user_uuid: UUID
    """Required. The user UUID to search projects for."""

    def to_condition(self) -> QueryCondition:
        """Convert scope to a query condition for AssocGroupUserRow.

        This will be used in a JOIN query with GroupRow.
        """
        user_uuid = self.user_uuid

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return AssocGroupUserRow.user_id == user_uuid

        return inner

    @property
    def existence_checks(self) -> Sequence[ExistenceCheck[UUID]]:
        """Return existence checks for scope validation.

        Note: User existence is typically already validated by auth layer.
        """
        return []



from __future__ import annotations

from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec

from .actions import (
    SearchDeploymentHistoryAction,
    SearchDeploymentHistoryActionResult,
    SearchDeploymentScopedHistoryAction,
    SearchDeploymentScopedHistoryActionResult,
    SearchRouteHistoryAction,
    SearchRouteHistoryActionResult,
    SearchRouteScopedHistoryAction,
    SearchRouteScopedHistoryActionResult,
    SearchSessionHistoryAction,
    SearchSessionHistoryActionResult,
    SearchSessionScopedHistoryAction,
    SearchSessionScopedHistoryActionResult,
)
from .service import SchedulingHistoryService


class SchedulingHistoryProcessors(AbstractProcessorPackage):
    """Processor package for scheduling history operations."""

    # Admin processors
    search_session_history: ActionProcessor[
        SearchSessionHistoryAction, SearchSessionHistoryActionResult
    ]
    search_deployment_history: ActionProcessor[
        SearchDeploymentHistoryAction, SearchDeploymentHistoryActionResult
    ]
    search_route_history: ActionProcessor[SearchRouteHistoryAction, SearchRouteHistoryActionResult]

    # Scoped processors (added in 26.2.0)
    search_session_scoped_history: ActionProcessor[
        SearchSessionScopedHistoryAction, SearchSessionScopedHistoryActionResult
    ]
    search_deployment_scoped_history: ActionProcessor[
        SearchDeploymentScopedHistoryAction, SearchDeploymentScopedHistoryActionResult
    ]
    search_route_scoped_history: ActionProcessor[
        SearchRouteScopedHistoryAction, SearchRouteScopedHistoryActionResult
    ]

    def __init__(
        self, service: SchedulingHistoryService, action_monitors: list[ActionMonitor]
    ) -> None:
        # Admin processors
        self.search_session_history = ActionProcessor(
            service.search_session_history, action_monitors
        )
        self.search_deployment_history = ActionProcessor(
            service.search_deployment_history, action_monitors
        )
        self.search_route_history = ActionProcessor(service.search_route_history, action_monitors)

        # Scoped processors (added in 26.2.0)
        self.search_session_scoped_history = ActionProcessor(
            service.search_session_scoped_history, action_monitors
        )
        self.search_deployment_scoped_history = ActionProcessor(
            service.search_deployment_scoped_history, action_monitors
        )
        self.search_route_scoped_history = ActionProcessor(
            service.search_route_scoped_history, action_monitors
        )

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            # Admin actions
            SearchSessionHistoryAction.spec(),
            SearchDeploymentHistoryAction.spec(),
            SearchRouteHistoryAction.spec(),
            # Scoped actions (added in 26.2.0)
            SearchSessionScopedHistoryAction.spec(),
            SearchDeploymentScopedHistoryAction.spec(),
            SearchRouteScopedHistoryAction.spec(),
        ]



from __future__ import annotations

import asyncio
import logging
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING

import aiofiles
import click

from ai.backend.common.json import pretty_json_str
from ai.backend.manager.api.gql.schema import schema as strawberry_schema
from ai.backend.manager.api.gql_legacy.schema import graphene_schema
from ai.backend.manager.openapi import generate

if TYPE_CHECKING:
    from .context import CLIContext

log = logging.getLogger(__spec__.name)


@click.group()
def cli() -> None:
    pass


async def generate_graphene_gql_schema(output_path: Path | None) -> None:
    if output_path is None:
        log.info("======== Graphene GraphQL API Schema ========")
        print(str(graphene_schema))
    else:
        async with aiofiles.open(output_path, "w") as fw:
            await fw.write(str(graphene_schema))


async def generate_strawberry_gql_schema(output_path: Path | None) -> None:
    if output_path is None:
        log.info("======== Strawberry GraphQL API Schema ========")
        print(strawberry_schema.as_str())
    else:
        async with aiofiles.open(output_path, "w") as fw:
            await fw.write(strawberry_schema.as_str())


def generate_supergraph_schema(
    supergraph_config_path: str | Path,
    output_dir: str | Path,
) -> None:
    """
    Post-processes GraphQL schema and generates supergraph.

    Args:
        schema_file_path: Path to the main schema file (e.g., schema2.graphql)
        supergraph_config_path: Path to the supergraph.yaml configuration file
        output_dir: Output directory (defaults to same directory as schema file)
    """
    log.info("Generating supergraph schema...")

    config_path = Path(supergraph_config_path)

    # Generate supergraph
    supergraph_path = Path(output_dir) / "supergraph.graphql"
    # Find the project root directory (where the supergraph.yaml paths are relative to)
    project_root = config_path
    while project_root.parent != project_root:
        if (project_root / "pyproject.toml").exists() or (project_root / ".git").exists():
            break
        project_root = project_root.parent

    result = subprocess.run(
        ["rover", "supergraph", "compose", "--config", str(config_path)],
        cwd=project_root,
        capture_output=True,
        text=True,
        check=True,
    )

    supergraph_path.write_text(result.stdout, encoding="utf-8")
    print(f"Supergraph generated at: {supergraph_path}")


@cli.command()
@click.pass_obj
@click.option(
    "--config",
    "-c",
    required=True,
    type=click.Path(exists=True, dir_okay=False, readable=True),
    help="Path to the supergraph.yaml configuration file",
)
@click.option(
    "--output-dir",
    "-o",
    default="docs/manager/graphql-reference",
    type=click.Path(file_okay=False, writable=True),
    help="Output directory for supergraph.graphql (default: same as schema file directory)",
)
def generate_supergraph(_cli_ctx: CLIContext, config: Path, output_dir: Path) -> None:
    """Post-process GraphQL schema and generate supergraph."""
    try:
        generate_supergraph_schema(
            supergraph_config_path=config,
            output_dir=output_dir,
        )
        click.echo("✅ Supergraph generation completed successfully!")
    except FileNotFoundError as e:
        click.echo(f"❌ Error: {e}", err=True)
        raise click.Abort() from e
    except subprocess.CalledProcessError as e:
        click.echo(f"❌ Rover command failed: {e}", err=True)
        if hasattr(e, "stderr") and e.stderr:
            click.echo(f"Error details: {e.stderr}", err=True)
        raise click.Abort() from e
    except Exception as e:
        click.echo(f"❌ Unexpected error: {e}", err=True)
        raise click.Abort() from e


@cli.command()
@click.pass_obj
@click.option(
    "--output",
    "-o",
    default="-",
    type=click.Path(dir_okay=False, writable=True, path_type=Path),
    help="Output file path (default: stdout)",
)
@click.option(
    "--v2",
    is_flag=True,
    default=False,  # TODO: Set default to True after v2 migration is complete
    help="Generate strawberry based v2 GraphQL schema (default: False)",
)
def dump_gql_schema(_cli_ctx: CLIContext, output: Path, v2: bool) -> None:
    """
    Generates GraphQL schema of Backend.AI API.
    """
    if v2:
        asyncio.run(generate_strawberry_gql_schema(output))
    else:
        asyncio.run(generate_graphene_gql_schema(output))


@cli.command()
@click.pass_obj
@click.option(
    "--output",
    "-o",
    default=None,
    type=click.Path(dir_okay=False, writable=True, path_type=Path),
    help="Output file path (default: stdout)",
)
def dump_openapi(_cli_ctx: CLIContext, output: Path | None) -> None:
    """
    Generates OpenAPI specification of Backend.AI API.
    """
    openapi = asyncio.run(generate())
    if output is None:
        print(pretty_json_str(openapi))
    else:
        with output.open(mode="w") as fw:
            fw.write(pretty_json_str(openapi))



"""Access token GraphQL types for model deployment."""

from __future__ import annotations

from collections.abc import Iterable
from datetime import datetime
from typing import Any, Self, override
from uuid import UUID

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.manager.api.gql.base import DateTimeFilter, OrderDirection, StringFilter
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy, StrawberryGQLContext
from ai.backend.manager.data.deployment.access_token import ModelDeploymentAccessTokenCreator
from ai.backend.manager.data.deployment.types import (
    AccessTokenOrderField,
    ModelDeploymentAccessTokenData,
)
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder
from ai.backend.manager.repositories.deployment.options import (
    AccessTokenConditions,
    AccessTokenOrders,
)


@strawberry.input(description="Added in 25.16.0")
class AccessTokenFilter(GQLFilter):
    """Filter for access tokens."""

    token: StringFilter | None = None
    valid_until: DateTimeFilter | None = None
    created_at: DateTimeFilter | None = None

    @override
    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from this filter."""
        conditions: list[QueryCondition] = []

        if self.token:
            if self.token.equals:
                conditions.append(AccessTokenConditions.by_token_equals(self.token.equals))
            elif self.token.contains:
                conditions.append(AccessTokenConditions.by_token_contains(self.token.contains))

        if self.valid_until:
            condition = self.valid_until.build_query_condition(
                before_factory=AccessTokenConditions.by_valid_until_before,
                after_factory=AccessTokenConditions.by_valid_until_after,
                equals_factory=AccessTokenConditions.by_valid_until_equals,
            )
            if condition:
                conditions.append(condition)

        if self.created_at:
            condition = self.created_at.build_query_condition(
                before_factory=AccessTokenConditions.by_created_at_before,
                after_factory=AccessTokenConditions.by_created_at_after,
                equals_factory=AccessTokenConditions.by_created_at_equals,
            )
            if condition:
                conditions.append(condition)

        return conditions


@strawberry.input(description="Added in 25.16.0")
class AccessTokenOrderBy(GQLOrderBy):
    field: AccessTokenOrderField
    direction: OrderDirection = OrderDirection.DESC

    @override
    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder."""
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case AccessTokenOrderField.CREATED_AT:
                return AccessTokenOrders.created_at(ascending)


@strawberry.type
class AccessToken(Node):
    id: NodeID[str]
    token: str = strawberry.field(description="Added in 25.16.0: The access token.")
    created_at: datetime = strawberry.field(
        description="Added in 25.16.0: The creation timestamp of the access token."
    )
    valid_until: datetime = strawberry.field(
        description="Added in 25.16.0: The expiration timestamp of the access token."
    )

    @classmethod
    async def resolve_nodes(  # type: ignore[override]  # Strawberry Node uses AwaitableOrValue overloads incompatible with async def
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.access_token_loader.load_many([
            UUID(nid) for nid in node_ids
        ])
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @classmethod
    def from_dataclass(cls, data: ModelDeploymentAccessTokenData) -> Self:
        return cls(
            id=ID(str(data.id)),
            token=data.token,
            created_at=data.created_at,
            valid_until=data.valid_until,
        )


AccessTokenEdge = Edge[AccessToken]


@strawberry.type(description="Added in 25.16.0")
class AccessTokenConnection(Connection[AccessToken]):
    count: int

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count


@strawberry.input
class CreateAccessTokenInput:
    model_deployment_id: ID = strawberry.field(
        description="Added in 25.16.0: The ID of the model deployment for which the access token is created."
    )
    valid_until: datetime = strawberry.field(
        description="Added in 25.16.0: The expiration timestamp of the access token."
    )

    def to_creator(self) -> ModelDeploymentAccessTokenCreator:
        return ModelDeploymentAccessTokenCreator(
            model_deployment_id=UUID(self.model_deployment_id),
            valid_until=self.valid_until,
        )


@strawberry.type
class CreateAccessTokenPayload:
    access_token: AccessToken



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.deployment.access_token import ModelDeploymentAccessTokenCreator
from ai.backend.manager.data.deployment.types import ModelDeploymentAccessTokenData
from ai.backend.manager.services.deployment.actions.access_token.base import (
    DeploymentAccessTokenBaseAction,
)


@dataclass
class CreateAccessTokenAction(DeploymentAccessTokenBaseAction):
    creator: ModelDeploymentAccessTokenCreator

    @override
    def entity_id(self) -> str | None:
        return str(self.creator.model_deployment_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class CreateAccessTokenActionResult(BaseActionResult):
    data: ModelDeploymentAccessTokenData

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id)



"""Configuration-related data types."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from ai.backend.common.types import AgentSelectionStrategy, ClusterSSHPortMapping


@dataclass
class SchedulingConfig:
    """Configuration needed for scheduling decisions."""

    max_container_count_per_agent: int | None
    enforce_spreading_endpoint_replica: bool


@dataclass
class ScalingGroupInfo:
    """Scaling group configuration for scheduling."""

    scheduler_name: str
    agent_selection_strategy: AgentSelectionStrategy


@dataclass
class NetworkSetup:
    """Network configuration for a session."""

    network_name: str | None = None
    network_config: dict[str, Any] = field(default_factory=dict)
    cluster_ssh_port_mapping: ClusterSSHPortMapping | None = None



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.resource_preset.types import ResourcePresetData
from ai.backend.manager.models.resource_preset import ResourcePresetRow
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.services.resource_preset.actions.base import ResourcePresetAction


@dataclass
class CreateResourcePresetAction(ResourcePresetAction):
    creator: Creator[ResourcePresetRow]

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class CreateResourcePresetActionResult(BaseActionResult):
    resource_preset: ResourcePresetData

    @override
    def entity_id(self) -> str | None:
        return str(self.resource_preset.id)


# TODO: Create exceptions.



"""
REST API handler for Session CSV export.
"""

from __future__ import annotations

from http import HTTPStatus

from aiohttp import web

from ai.backend.common.api_handlers import (
    APIStreamResponse,
    BodyParam,
    HeaderParam,
    stream_api_handler,
)
from ai.backend.common.contexts.user import current_user
from ai.backend.common.dto.manager.export import SessionExportCSVRequest
from ai.backend.manager.api.auth import auth_required_for_method
from ai.backend.manager.dto.context import ExportCtx, ProcessorsCtx
from ai.backend.manager.dto.export import ExportFilenameHeader
from ai.backend.manager.exporter.csv import CSVExporter
from ai.backend.manager.exporter.stream import CSVExportStreamReader
from ai.backend.manager.repositories.base.export import ExportDataStream
from ai.backend.manager.services.export.actions import ExportSessionsCSVAction

from .adapter import ExportAdapter

__all__ = ("SessionExportHandler",)

SESSIONS_REPORT_KEY = "sessions"


class SessionExportHandler:
    """REST API handler for session CSV export operations."""

    def __init__(self) -> None:
        self._adapter = ExportAdapter()

    @auth_required_for_method
    @stream_api_handler
    async def export_csv(
        self,
        body: BodyParam[SessionExportCSVRequest],
        header: HeaderParam[ExportFilenameHeader],
        processors_ctx: ProcessorsCtx,
        export_ctx: ExportCtx,
    ) -> APIStreamResponse:
        """Export session data as CSV.

        POST /export/sessions/csv
        """
        me = current_user()
        if me is None or not me.is_superadmin:
            raise web.HTTPForbidden(reason="Only superadmin can export data.")

        # Get report definition from repository
        report = export_ctx.repository.get_report(SESSIONS_REPORT_KEY)

        # Build query using adapter
        query = self._adapter.build_session_query(
            report=report,
            fields=body.parsed.fields,
            filter=body.parsed.filter,
            order=body.parsed.order,
            max_rows=export_ctx.config.max_rows,
            statement_timeout_sec=export_ctx.config.statement_timeout_sec,
        )

        # Create action with pre-built query
        action = ExportSessionsCSVAction(
            query=query,
            encoding=body.parsed.encoding,
            filename=header.parsed.filename,
        )

        # Call service through processors
        action_result = (
            await processors_ctx.processors.export.export_sessions_csv.wait_for_complete(action)
        )

        # Create CSV exporter and stream reader
        data_stream = ExportDataStream(
            field_names=action_result.field_names,
            reader=action_result.row_iterator,
        )
        exporter = CSVExporter(data_stream, encoding=action_result.encoding)
        stream_reader = CSVExportStreamReader(exporter)

        # Build response headers
        headers = {
            "Content-Disposition": f'attachment; filename="{action_result.filename}"',
            "Content-Type": f"text/csv; charset={action_result.encoding}",
        }

        return APIStreamResponse(body=stream_reader, status=HTTPStatus.OK, headers=headers)



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.agent.types import AgentDetailData
from ai.backend.manager.repositories.base import BatchQuerier

from .base import AgentAction


@dataclass
class SearchAgentsAction(AgentAction):
    querier: BatchQuerier

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class SearchAgentsActionResult(BaseActionResult):
    """Result of searching agents with their permissions."""

    agents: list[AgentDetailData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

import uuid
from collections.abc import Sequence

from ai.backend.manager.data.vfs_storage.types import VFSStorageData
from ai.backend.manager.repositories.base import BatchQuerier, OffsetPagination
from ai.backend.manager.repositories.vfs_storage.options import VFSStorageConditions
from ai.backend.manager.services.vfs_storage.actions.search import SearchVFSStoragesAction
from ai.backend.manager.services.vfs_storage.processors import VFSStorageProcessors


async def load_vfs_storages_by_ids(
    processor: VFSStorageProcessors,
    storage_ids: Sequence[uuid.UUID],
) -> list[VFSStorageData | None]:
    """Batch load VFS storages by their IDs.

    Args:
        processor: The VFS storage processor.
        storage_ids: Sequence of storage UUIDs to load.

    Returns:
        List of VFSStorageData (or None if not found) in the same order as storage_ids.
    """
    if not storage_ids:
        return []

    querier = BatchQuerier(
        pagination=OffsetPagination(limit=len(storage_ids)),
        conditions=[VFSStorageConditions.by_ids(storage_ids)],
    )

    action_result = await processor.search_vfs_storages.wait_for_complete(
        SearchVFSStoragesAction(querier=querier)
    )

    storage_map = {storage.id: storage for storage in action_result.storages}
    return [storage_map.get(storage_id) for storage_id in storage_ids]



from collections.abc import Collection
from typing import TypeVar

from sqlalchemy import inspect
from sqlalchemy.dialects.postgresql import insert as pg_insert
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.manager.models.base import Base

TRow = TypeVar("TRow", bound=Base)


async def insert_on_conflict_do_nothing[TRow: Base](
    db_sess: SASession,
    row: TRow,
) -> TRow:
    """Insert the given row, ignoring conflicts."""
    mapper = inspect(type(row))
    column_keys = {c.key for c in mapper.columns}
    values = {k: v for k, v in row.__dict__.items() if k in column_keys}
    stmt = pg_insert(type(row)).values(**values).on_conflict_do_nothing()
    await db_sess.execute(stmt)
    await db_sess.flush()
    return row


async def bulk_insert_on_conflict_do_nothing[TRow: Base](
    db_sess: SASession,
    rows: Collection[TRow],
) -> None:
    """Insert the given rows, ignoring conflicts."""
    if not rows:
        return
    row_cls = type(next(iter(rows)))
    mapper = inspect(row_cls)
    column_keys = {c.key for c in mapper.columns}
    values_list = [{k: v for k, v in row.__dict__.items() if k in column_keys} for row in rows]
    stmt = pg_insert(row_cls).values(values_list).on_conflict_do_nothing()
    await db_sess.execute(stmt)
    await db_sess.flush()



"""
Configuration migration CLI commands for Backend.AI Manager.
"""

import pathlib

import click

from ai.backend.common.configs.migration.cli import run_sd_migration, sd_migration_options
from ai.backend.manager.config.sd_migration import MANAGER_MAPPING_RULES


@click.group()
def cli() -> None:
    """Migrate legacy configuration fields."""
    pass


@cli.command(name="service-discovery")
@sd_migration_options
def service_discovery(
    input_path: pathlib.Path,
    output_path: pathlib.Path | None,
    append: bool,
    dry_run: bool,
    force: bool,
) -> None:
    """Migrate address/port fields to [service-discovery.endpoints]."""
    run_sd_migration(
        input_path=input_path,
        output_path=output_path,
        append=append,
        dry_run=dry_run,
        force=force,
        component_name="manager",
        mapping_rules=MANAGER_MAPPING_RULES,
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.models.scaling_group import ScalingGroupForKeypairsRow
from ai.backend.manager.repositories.base.purger import BatchPurger

from .keypair_base import ScalingGroupKeypairAction


@dataclass
class DisassociateScalingGroupWithKeypairsAction(ScalingGroupKeypairAction):
    """Action to disassociate a scaling group from multiple keypairs."""

    purger: BatchPurger[ScalingGroupForKeypairsRow]

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class DisassociateScalingGroupWithKeypairsActionResult(BaseActionResult):
    """Result of disassociating a scaling group from keypairs."""

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

import uuid
from collections.abc import Iterable
from typing import Self

import strawberry
from strawberry import ID, UNSET, Info
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.manager.api.gql.base import encode_cursor
from ai.backend.manager.data.vfs_storage.types import VFSStorageData
from ai.backend.manager.models.vfs_storage import VFSStorageRow
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.repositories.vfs_storage import VFSStorageCreatorSpec
from ai.backend.manager.repositories.vfs_storage.updaters import VFSStorageUpdaterSpec
from ai.backend.manager.services.vfs_storage.actions.create import CreateVFSStorageAction
from ai.backend.manager.services.vfs_storage.actions.delete import DeleteVFSStorageAction
from ai.backend.manager.services.vfs_storage.actions.get import GetVFSStorageAction
from ai.backend.manager.services.vfs_storage.actions.list import ListVFSStorageAction
from ai.backend.manager.services.vfs_storage.actions.update import UpdateVFSStorageAction
from ai.backend.manager.types import OptionalState

from .types import StrawberryGQLContext


@strawberry.type(description="Added in 25.16.0. VFS Storage configuration")
class VFSStorage(Node):
    id: NodeID[str]
    name: str
    host: str
    base_path: str

    @classmethod
    async def resolve_nodes(  # type: ignore[override]  # Strawberry Node uses AwaitableOrValue overloads incompatible with async def
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.vfs_storage_loader.load_many([
            uuid.UUID(nid) for nid in node_ids
        ])
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @classmethod
    def from_dataclass(cls, data: VFSStorageData) -> Self:
        return cls(
            id=ID(str(data.id)),
            name=data.name,
            host=data.host,
            base_path=str(data.base_path),
        )


VFSStorageEdge = Edge[VFSStorage]


@strawberry.type(description="Added in 25.16.0. VFS Storage connection")
class VFSStorageConnection(Connection[VFSStorage]):
    @strawberry.field
    def count(self) -> int:
        return len(self.edges)


@strawberry.field(description="Added in 25.16.0. Get a VFS storage by ID")  # type: ignore[misc]
async def vfs_storage(id: ID, info: Info[StrawberryGQLContext]) -> VFSStorage | None:
    processors = info.context.processors
    action_result = await processors.vfs_storage.get.wait_for_complete(
        GetVFSStorageAction(storage_id=uuid.UUID(id))
    )
    return VFSStorage.from_dataclass(action_result.result)


@strawberry.field(description="Added in 25.16.0. List all VFS storages")  # type: ignore[misc]
async def vfs_storages(
    info: Info[StrawberryGQLContext],
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> VFSStorageConnection | None:
    processors = info.context.processors

    action_result = await processors.vfs_storage.list_storages.wait_for_complete(
        ListVFSStorageAction()
    )

    nodes = [VFSStorage.from_dataclass(data) for data in action_result.data]
    edges = [VFSStorageEdge(node=node, cursor=encode_cursor(node.id)) for node in nodes]

    return VFSStorageConnection(
        edges=edges,
        page_info=strawberry.relay.PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
    )


@strawberry.input(description="Added in 25.16.0. Input for creating VFS storage")
class CreateVFSStorageInput:
    name: str
    host: str
    base_path: str

    def to_creator(self) -> Creator[VFSStorageRow]:
        return Creator(
            spec=VFSStorageCreatorSpec(
                name=self.name,
                host=self.host,
                base_path=self.base_path,
            )
        )


@strawberry.input(description="Added in 25.16.0. Input for updating VFS storage")
class UpdateVFSStorageInput:
    id: ID
    name: str | None = UNSET
    host: str | None = UNSET
    base_path: str | None = UNSET

    def to_updater(self) -> Updater[VFSStorageRow]:
        spec = VFSStorageUpdaterSpec(
            name=OptionalState[str].from_graphql(self.name),
            host=OptionalState[str].from_graphql(self.host),
            base_path=OptionalState[str].from_graphql(self.base_path),
        )
        return Updater(spec=spec, pk_value=uuid.UUID(self.id))


@strawberry.input(description="Added in 25.16.0. Input for deleting VFS storage")
class DeleteVFSStorageInput:
    id: ID


@strawberry.type(description="Added in 25.16.0. Payload for creating VFS storage")
class CreateVFSStoragePayload:
    vfs_storage: VFSStorage


@strawberry.type(description="Added in 25.16.0. Payload for updating VFS storage")
class UpdateVFSStoragePayload:
    vfs_storage: VFSStorage


@strawberry.type(description="Added in 25.16.0. Payload for deleting VFS storage")
class DeleteVFSStoragePayload:
    id: ID


@strawberry.mutation(  # type: ignore[misc]
    name="createVFSStorage", description="Added in 25.16.0. Create a new VFS storage"
)
async def create_vfs_storage(
    input: CreateVFSStorageInput, info: Info[StrawberryGQLContext]
) -> CreateVFSStoragePayload:
    processors = info.context.processors

    action_result = await processors.vfs_storage.create.wait_for_complete(
        CreateVFSStorageAction(
            creator=input.to_creator(),
        )
    )

    return CreateVFSStoragePayload(vfs_storage=VFSStorage.from_dataclass(action_result.result))


@strawberry.mutation(  # type: ignore[misc]
    name="updateVFSStorage", description="Added in 25.16.0. Update an existing VFS storage"
)
async def update_vfs_storage(
    input: UpdateVFSStorageInput, info: Info[StrawberryGQLContext]
) -> UpdateVFSStoragePayload:
    processors = info.context.processors

    action_result = await processors.vfs_storage.update.wait_for_complete(
        UpdateVFSStorageAction(
            updater=input.to_updater(),
        )
    )

    return UpdateVFSStoragePayload(vfs_storage=VFSStorage.from_dataclass(action_result.result))


@strawberry.mutation(name="deleteVFSStorage", description="Added in 25.16.0. Delete a VFS storage")  # type: ignore[misc]
async def delete_vfs_storage(
    input: DeleteVFSStorageInput, info: Info[StrawberryGQLContext]
) -> DeleteVFSStoragePayload:
    processors = info.context.processors

    action_result = await processors.vfs_storage.delete.wait_for_complete(
        DeleteVFSStorageAction(
            storage_id=uuid.UUID(input.id),
        )
    )

    return DeleteVFSStoragePayload(id=ID(str(action_result.deleted_storage_id)))



from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience import (
    MetricArgs,
    MetricPolicy,
    Resilience,
    RetryArgs,
    RetryPolicy,
)
from ai.backend.common.resilience.policies.retry import BackoffStrategy
from ai.backend.manager.data.deployment.types import (
    DeploymentHistoryListResult,
    RouteHistoryListResult,
)
from ai.backend.manager.data.kernel.types import (
    KernelSchedulingHistoryListResult,
)
from ai.backend.manager.data.session.types import (
    SessionSchedulingHistoryListResult,
)
from ai.backend.manager.repositories.base import BatchQuerier

from .db_source import SchedulingHistoryDBSource
from .types import (
    DeploymentHistorySearchScope,
    RouteHistorySearchScope,
    SessionSchedulingHistorySearchScope,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

__all__ = ("SchedulingHistoryRepository",)


scheduling_history_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.SCHEDULING_HISTORY_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
            )
        ),
    ]
)


class SchedulingHistoryRepository:
    """Repository for scheduling history data access (read-only)."""

    _db_source: SchedulingHistoryDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = SchedulingHistoryDBSource(db)

    # ========== Session History (Admin) ==========

    @scheduling_history_repository_resilience.apply()
    async def search_session_history(
        self,
        querier: BatchQuerier,
    ) -> SessionSchedulingHistoryListResult:
        """Search session scheduling history with pagination (admin API)."""
        return await self._db_source.search_session_history(querier)

    # ========== Session History (Scoped) ==========

    @scheduling_history_repository_resilience.apply()
    async def search_session_scoped_history(
        self,
        querier: BatchQuerier,
        scope: SessionSchedulingHistorySearchScope,
    ) -> SessionSchedulingHistoryListResult:
        """Search session scheduling history within scope."""
        return await self._db_source.search_session_scoped_history(querier, scope)

    # ========== Kernel History ==========

    @scheduling_history_repository_resilience.apply()
    async def search_kernel_history(
        self,
        querier: BatchQuerier,
    ) -> KernelSchedulingHistoryListResult:
        """Search kernel scheduling history with pagination."""
        return await self._db_source.search_kernel_history(querier)

    # ========== Deployment History (Admin) ==========

    @scheduling_history_repository_resilience.apply()
    async def search_deployment_history(
        self,
        querier: BatchQuerier,
    ) -> DeploymentHistoryListResult:
        """Search deployment history with pagination (admin API)."""
        return await self._db_source.search_deployment_history(querier)

    # ========== Deployment History (Scoped) ==========

    @scheduling_history_repository_resilience.apply()
    async def search_deployment_scoped_history(
        self,
        querier: BatchQuerier,
        scope: DeploymentHistorySearchScope,
    ) -> DeploymentHistoryListResult:
        """Search deployment history within scope."""
        return await self._db_source.search_deployment_scoped_history(querier, scope)

    # ========== Route History (Admin) ==========

    @scheduling_history_repository_resilience.apply()
    async def search_route_history(
        self,
        querier: BatchQuerier,
    ) -> RouteHistoryListResult:
        """Search route history with pagination (admin API)."""
        return await self._db_source.search_route_history(querier)

    # ========== Route History (Scoped) ==========

    @scheduling_history_repository_resilience.apply()
    async def search_route_scoped_history(
        self,
        querier: BatchQuerier,
        scope: RouteHistorySearchScope,
    ) -> RouteHistoryListResult:
        """Search route history within scope."""
        return await self._db_source.search_route_scoped_history(querier, scope)



from __future__ import annotations

import uuid

import sqlalchemy as sa
from sqlalchemy.orm import selectinload

from ai.backend.common.data.artifact.types import ArtifactRegistryType
from ai.backend.manager.data.artifact_registries.types import (
    ArtifactRegistryCreatorMeta,
    ArtifactRegistryModifierMeta,
)
from ai.backend.manager.data.reservoir_registry.types import (
    ReservoirRegistryData,
    ReservoirRegistryListResult,
)
from ai.backend.manager.errors.artifact import ArtifactNotFoundError
from ai.backend.manager.errors.artifact_registry import ArtifactRegistryNotFoundError
from ai.backend.manager.models.artifact import ArtifactRow
from ai.backend.manager.models.artifact_registries import ArtifactRegistryRow
from ai.backend.manager.models.reservoir_registry import ReservoirRegistryRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.base import BatchQuerier, execute_batch_querier
from ai.backend.manager.repositories.base.creator import Creator, execute_creator
from ai.backend.manager.repositories.base.updater import Updater, execute_updater


class ReservoirDBSource:
    """Database source for reservoir registry operations."""

    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def get_reservoir_registry_data_by_id(
        self, reservoir_id: uuid.UUID
    ) -> ReservoirRegistryData:
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            result = await db_sess.execute(
                sa.select(ReservoirRegistryRow)
                .where(ReservoirRegistryRow.id == reservoir_id)
                .options(selectinload(ReservoirRegistryRow.meta))
            )
            row = result.scalar_one_or_none()
            if row is None:
                raise ArtifactRegistryNotFoundError(f"Reservoir with ID {reservoir_id} not found")
            return row.to_dataclass()

    async def get_registries_by_ids(
        self, reservoir_ids: list[uuid.UUID]
    ) -> list[ReservoirRegistryData]:
        """
        Get multiple Reservoir registry entries by their IDs in a single query.
        """
        async with self._db.begin_readonly_session_read_committed() as db_session:
            result = await db_session.execute(
                sa.select(ReservoirRegistryRow)
                .where(ReservoirRegistryRow.id.in_(reservoir_ids))
                .options(selectinload(ReservoirRegistryRow.meta))
            )
            rows = result.scalars().all()
            return [row.to_dataclass() for row in rows]

    async def get_registry_data_by_name(self, name: str) -> ReservoirRegistryData:
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            result = await db_sess.execute(
                sa.select(ArtifactRegistryRow)
                .where(ArtifactRegistryRow.name == name)
                .options(
                    selectinload(ArtifactRegistryRow.reservoir_registries).selectinload(
                        ReservoirRegistryRow.meta
                    )
                )
            )
            row = result.scalar_one_or_none()
            if row is None:
                raise ArtifactRegistryNotFoundError(f"Registry with name {name} not found")
            if row.reservoir_registries is None:
                raise ArtifactRegistryNotFoundError(
                    f"Reservoir registry not found for registry {name}"
                )
            return row.reservoir_registries.to_dataclass()

    async def get_registry_data_by_artifact_id(
        self, artifact_id: uuid.UUID
    ) -> ReservoirRegistryData:
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            result = await db_sess.execute(
                sa.select(ArtifactRow)
                .where(ArtifactRow.id == artifact_id)
                .options(
                    selectinload(ArtifactRow.reservoir_registry).selectinload(
                        ReservoirRegistryRow.meta
                    ),
                )
            )
            row = result.scalar_one_or_none()
            if row is None:
                raise ArtifactNotFoundError(f"Artifact with ID {artifact_id} not found")
            if row.reservoir_registry is None:
                raise ArtifactRegistryNotFoundError(
                    f"Reservoir registry not found for artifact {artifact_id}"
                )
            return row.reservoir_registry.to_dataclass()

    async def create(
        self, creator: Creator[ReservoirRegistryRow], meta: ArtifactRegistryCreatorMeta
    ) -> ReservoirRegistryData:
        """
        Create a new Reservoir entry.
        """
        async with self._db.begin_session() as db:
            creator_result = await execute_creator(db, creator)
            new_row = creator_result.row

            reg_insert = sa.insert(ArtifactRegistryRow).values(
                name=meta.name,
                registry_id=new_row.id,
                type=ArtifactRegistryType.RESERVOIR,
            )
            await db.execute(reg_insert)

            stmt = (
                sa.select(ReservoirRegistryRow)
                .where(ReservoirRegistryRow.id == new_row.id)
                .options(selectinload(ReservoirRegistryRow.meta))
            )

            row: ReservoirRegistryRow | None = (await db.execute(stmt)).scalar_one_or_none()
            if row is None:
                raise ArtifactRegistryNotFoundError(f"Registry with ID {new_row.id} not found")

            return row.to_dataclass()

    async def update(
        self,
        updater: Updater[ReservoirRegistryRow],
        meta: ArtifactRegistryModifierMeta,
    ) -> ReservoirRegistryData:
        """
        Update an existing Reservoir entry in the database.
        """
        async with self._db.begin_session() as db_session:
            result = await execute_updater(db_session, updater)
            if result is None:
                raise ArtifactRegistryNotFoundError(
                    f"Reservoir registry with ID {updater.pk_value} not found"
                )
            updated_row_id = result.row.id

            if (name := meta.name.optional_value()) is not None:
                await db_session.execute(
                    sa.update(ArtifactRegistryRow)
                    .where(ArtifactRegistryRow.registry_id == updated_row_id)
                    .values(name=name)
                )

            # Reselect for the `selectinload`
            row = (
                await db_session.execute(
                    sa.select(ReservoirRegistryRow)
                    .where(ReservoirRegistryRow.id == updated_row_id)
                    .options(selectinload(ReservoirRegistryRow.meta))
                )
            ).scalar_one()

            return row.to_dataclass()

    async def delete(self, reservoir_id: uuid.UUID) -> uuid.UUID:
        """
        Delete an existing Reservoir entry from the database.
        """
        async with self._db.begin_session() as db_session:
            delete_query = (
                sa.delete(ReservoirRegistryRow)
                .where(ReservoirRegistryRow.id == reservoir_id)
                .returning(ReservoirRegistryRow.id)
            )
            result = await db_session.execute(delete_query)
            deleted_id = result.scalar()
            if deleted_id is None:
                raise ArtifactRegistryNotFoundError(
                    f"Reservoir registry with ID {reservoir_id} not found"
                )

            delete_meta_query = sa.delete(ArtifactRegistryRow).where(
                ArtifactRegistryRow.registry_id == reservoir_id
            )
            await db_session.execute(delete_meta_query)
            return deleted_id

    async def list_reservoir_registries(self) -> list[ReservoirRegistryData]:
        """
        List all Reservoir entries from the database.
        """
        async with self._db.begin_readonly_session_read_committed() as db_session:
            query = sa.select(ReservoirRegistryRow).options(selectinload(ReservoirRegistryRow.meta))
            result = await db_session.execute(query)
            rows = result.scalars().all()
            return [row.to_dataclass() for row in rows]

    async def search_registries(
        self,
        querier: BatchQuerier,
    ) -> ReservoirRegistryListResult:
        """Searches Reservoir registries with total count."""
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(ReservoirRegistryRow).options(selectinload(ReservoirRegistryRow.meta))

            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
            )

            items = [row.ReservoirRegistryRow.to_dataclass() for row in result.rows]

            return ReservoirRegistryListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, override

from ai.backend.manager.models.resource_policy import ProjectResourcePolicyRow
from ai.backend.manager.repositories.base.updater import UpdaterSpec
from ai.backend.manager.types import OptionalState


@dataclass
class ProjectResourcePolicyUpdaterSpec(UpdaterSpec[ProjectResourcePolicyRow]):
    """UpdaterSpec for project resource policy updates."""

    max_vfolder_count: OptionalState[int] = field(default_factory=OptionalState.nop)
    max_quota_scope_size: OptionalState[int] = field(default_factory=OptionalState.nop)
    max_vfolder_size: OptionalState[int] = field(default_factory=OptionalState.nop)
    max_network_count: OptionalState[int] = field(default_factory=OptionalState.nop)

    @property
    @override
    def row_class(self) -> type[ProjectResourcePolicyRow]:
        return ProjectResourcePolicyRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.max_vfolder_count.update_dict(to_update, "max_vfolder_count")
        self.max_quota_scope_size.update_dict(to_update, "max_quota_scope_size")
        self.max_vfolder_size.update_dict(to_update, "max_vfolder_size")
        self.max_network_count.update_dict(to_update, "max_network_count")
        return to_update



"""GraphQL types, filters, and inputs for scheduling history."""

from __future__ import annotations

from collections.abc import Iterable
from datetime import datetime
from enum import StrEnum
from typing import Self, override
from uuid import UUID

import strawberry
from strawberry import ID, Info
from strawberry.relay import Node, NodeID

from ai.backend.manager.api.gql.base import (
    DateTimeFilter,
    OrderDirection,
    StringFilter,
    UUIDFilter,
)
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy, StrawberryGQLContext
from ai.backend.manager.data.deployment.types import DeploymentHistoryData, RouteHistoryData
from ai.backend.manager.data.session.types import (
    SchedulingResult,
    SessionSchedulingHistoryData,
    SubStepResult,
)
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder
from ai.backend.manager.repositories.scheduling_history.options import (
    DeploymentHistoryConditions,
    DeploymentHistoryOrders,
    RouteHistoryConditions,
    RouteHistoryOrders,
    SessionSchedulingHistoryConditions,
    SessionSchedulingHistoryOrders,
)

__all__ = (
    # Enums
    "SchedulingResultGQL",
    "SessionSchedulingHistoryOrderField",
    "DeploymentHistoryOrderField",
    "RouteHistoryOrderField",
    # Types
    "SubStepResultGQL",
    "SessionSchedulingHistory",
    "DeploymentHistory",
    "RouteHistory",
    # Filters
    "SessionSchedulingHistoryFilter",
    "SessionSchedulingHistoryOrderBy",
    "DeploymentHistoryFilter",
    "DeploymentHistoryOrderBy",
    "RouteHistoryFilter",
    "RouteHistoryOrderBy",
    # Scope types (added in 26.2.0)
    "SessionScope",
    "DeploymentScope",
    "RouteScope",
)


# Enums


@strawberry.enum(name="SchedulingResult", description="Scheduling result status")
class SchedulingResultGQL(StrEnum):
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"
    STALE = "STALE"
    NEED_RETRY = "NEED_RETRY"
    EXPIRED = "EXPIRED"
    GIVE_UP = "GIVE_UP"
    SKIPPED = "SKIPPED"

    @classmethod
    def from_internal(cls, value: SchedulingResult) -> SchedulingResultGQL:
        match value:
            case SchedulingResult.SUCCESS:
                return cls.SUCCESS
            case SchedulingResult.FAILURE:
                return cls.FAILURE
            case SchedulingResult.STALE:
                return cls.STALE
            case SchedulingResult.NEED_RETRY:
                return cls.NEED_RETRY
            case SchedulingResult.EXPIRED:
                return cls.EXPIRED
            case SchedulingResult.GIVE_UP:
                return cls.GIVE_UP
            case SchedulingResult.SKIPPED:
                return cls.SKIPPED
            case _:
                raise ValueError(f"Unknown SchedulingResult: {value}")

    def to_internal(self) -> SchedulingResult:
        match self:
            case SchedulingResultGQL.SUCCESS:
                return SchedulingResult.SUCCESS
            case SchedulingResultGQL.FAILURE:
                return SchedulingResult.FAILURE
            case SchedulingResultGQL.STALE:
                return SchedulingResult.STALE
            case SchedulingResultGQL.NEED_RETRY:
                return SchedulingResult.NEED_RETRY
            case SchedulingResultGQL.EXPIRED:
                return SchedulingResult.EXPIRED
            case SchedulingResultGQL.GIVE_UP:
                return SchedulingResult.GIVE_UP
            case SchedulingResultGQL.SKIPPED:
                return SchedulingResult.SKIPPED


@strawberry.enum
class SessionSchedulingHistoryOrderField(StrEnum):
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"


@strawberry.enum
class DeploymentHistoryOrderField(StrEnum):
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"


@strawberry.enum
class RouteHistoryOrderField(StrEnum):
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"


# Types


@strawberry.type(description="Sub-step result in scheduling history")
class SubStepResultGQL:
    step: str
    result: SchedulingResultGQL
    error_code: str | None
    message: str | None
    started_at: datetime | None
    ended_at: datetime | None

    @classmethod
    def from_dataclass(cls, data: SubStepResult) -> Self:
        return cls(
            step=data.step,
            result=SchedulingResultGQL.from_internal(data.result),
            error_code=data.error_code,
            message=data.message,
            started_at=data.started_at,
            ended_at=data.ended_at,
        )


@strawberry.type(description="Session scheduling history record")
class SessionSchedulingHistory(Node):
    id: NodeID[str]
    _session_id: strawberry.Private[UUID]
    phase: str
    from_status: str | None
    to_status: str | None
    result: SchedulingResultGQL
    error_code: str | None
    message: str | None
    sub_steps: list[SubStepResultGQL]
    attempts: int
    created_at: datetime
    updated_at: datetime

    @strawberry.field(description="The session ID this history record belongs to.")  # type: ignore[misc]
    def session_id(self) -> ID:
        return ID(str(self._session_id))

    @classmethod
    async def resolve_nodes(  # type: ignore[override]  # Strawberry Node uses AwaitableOrValue overloads incompatible with async def
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.session_history_loader.load_many([
            UUID(nid) for nid in node_ids
        ])
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @classmethod
    def from_dataclass(cls, data: SessionSchedulingHistoryData) -> Self:
        return cls(
            id=ID(str(data.id)),
            _session_id=data.session_id,
            phase=data.phase,
            from_status=data.from_status.value if data.from_status else None,
            to_status=data.to_status.value if data.to_status else None,
            result=SchedulingResultGQL.from_internal(data.result),
            error_code=data.error_code,
            message=data.message,
            sub_steps=[SubStepResultGQL.from_dataclass(s) for s in data.sub_steps],
            attempts=data.attempts,
            created_at=data.created_at,
            updated_at=data.updated_at,
        )


@strawberry.type(description="Deployment history record")
class DeploymentHistory(Node):
    id: NodeID[str]
    _deployment_id: strawberry.Private[UUID]
    phase: str
    from_status: str | None
    to_status: str | None
    result: SchedulingResultGQL
    error_code: str | None
    message: str | None
    sub_steps: list[SubStepResultGQL]
    attempts: int
    created_at: datetime
    updated_at: datetime

    @strawberry.field(description="The deployment ID this history record belongs to.")  # type: ignore[misc]
    def deployment_id(self) -> ID:
        return ID(str(self._deployment_id))

    @classmethod
    async def resolve_nodes(  # type: ignore[override]  # Strawberry Node uses AwaitableOrValue overloads incompatible with async def
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.deployment_history_loader.load_many([
            UUID(nid) for nid in node_ids
        ])
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @classmethod
    def from_dataclass(cls, data: DeploymentHistoryData) -> Self:
        return cls(
            id=ID(str(data.id)),
            _deployment_id=data.deployment_id,
            phase=data.phase,
            from_status=data.from_status.value if data.from_status else None,
            to_status=data.to_status.value if data.to_status else None,
            result=SchedulingResultGQL.from_internal(data.result),
            error_code=data.error_code,
            message=data.message,
            sub_steps=[SubStepResultGQL.from_dataclass(s) for s in data.sub_steps],
            attempts=data.attempts,
            created_at=data.created_at,
            updated_at=data.updated_at,
        )


@strawberry.type(description="Route history record")
class RouteHistory(Node):
    id: NodeID[str]
    _route_id: strawberry.Private[UUID]
    _deployment_id: strawberry.Private[UUID]
    phase: str
    from_status: str | None
    to_status: str | None
    result: SchedulingResultGQL
    error_code: str | None
    message: str | None
    sub_steps: list[SubStepResultGQL]
    attempts: int
    created_at: datetime
    updated_at: datetime

    @strawberry.field(description="The route ID this history record belongs to.")  # type: ignore[misc]
    def route_id(self) -> ID:
        return ID(str(self._route_id))

    @strawberry.field(description="The deployment ID this route belongs to.")  # type: ignore[misc]
    def deployment_id(self) -> ID:
        return ID(str(self._deployment_id))

    @classmethod
    async def resolve_nodes(  # type: ignore[override]  # Strawberry Node uses AwaitableOrValue overloads incompatible with async def
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.route_history_loader.load_many([
            UUID(nid) for nid in node_ids
        ])
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @classmethod
    def from_dataclass(cls, data: RouteHistoryData) -> Self:
        return cls(
            id=ID(str(data.id)),
            _route_id=data.route_id,
            _deployment_id=data.deployment_id,
            phase=data.phase,
            from_status=data.from_status.value if data.from_status else None,
            to_status=data.to_status.value if data.to_status else None,
            result=SchedulingResultGQL.from_internal(data.result),
            error_code=data.error_code,
            message=data.message,
            sub_steps=[SubStepResultGQL.from_dataclass(s) for s in data.sub_steps],
            attempts=data.attempts,
            created_at=data.created_at,
            updated_at=data.updated_at,
        )


# Scope input types (added in 26.2.0)


@strawberry.input(description="Scope for session scheduling history query")
class SessionScope:
    """Scope for session-level scheduling history queries."""

    session_id: UUID = strawberry.field(description="Session ID to get history for")


@strawberry.input(description="Scope for deployment scheduling history query")
class DeploymentScope:
    """Scope for deployment-level scheduling history queries."""

    deployment_id: UUID = strawberry.field(description="Deployment ID to get history for")


@strawberry.input(description="Scope for route scheduling history query")
class RouteScope:
    """Scope for route-level scheduling history queries."""

    route_id: UUID = strawberry.field(description="Route ID to get history for")


# Filters


@strawberry.input(description="Filter for session scheduling history")
class SessionSchedulingHistoryFilter(GQLFilter):
    id: UUIDFilter | None = None
    session_id: UUIDFilter | None = None
    phase: StringFilter | None = None
    from_status: list[str] | None = None
    to_status: list[str] | None = None
    result: list[SchedulingResultGQL] | None = None
    error_code: StringFilter | None = None
    message: StringFilter | None = None
    created_at: DateTimeFilter | None = None
    updated_at: DateTimeFilter | None = None

    @override
    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from this filter."""
        conditions: list[QueryCondition] = []

        if self.id is not None:
            condition = self.id.build_query_condition(
                equals_factory=SessionSchedulingHistoryConditions.by_id_filter,
                in_factory=SessionSchedulingHistoryConditions.by_id_in,
            )
            if condition:
                conditions.append(condition)

        if self.session_id is not None:
            condition = self.session_id.build_query_condition(
                equals_factory=SessionSchedulingHistoryConditions.by_session_id_filter,
                in_factory=SessionSchedulingHistoryConditions.by_session_id_in,
            )
            if condition:
                conditions.append(condition)

        if self.phase is not None:
            condition = self.phase.build_query_condition(
                contains_factory=SessionSchedulingHistoryConditions.by_phase_contains,
                equals_factory=SessionSchedulingHistoryConditions.by_phase_equals,
                starts_with_factory=SessionSchedulingHistoryConditions.by_phase_starts_with,
                ends_with_factory=SessionSchedulingHistoryConditions.by_phase_ends_with,
            )
            if condition:
                conditions.append(condition)

        if self.from_status is not None and len(self.from_status) > 0:
            conditions.append(SessionSchedulingHistoryConditions.by_from_statuses(self.from_status))

        if self.to_status is not None and len(self.to_status) > 0:
            conditions.append(SessionSchedulingHistoryConditions.by_to_statuses(self.to_status))

        if self.result is not None and len(self.result) > 0:
            conditions.append(
                SessionSchedulingHistoryConditions.by_results([
                    r.to_internal() for r in self.result
                ])
            )

        if self.error_code is not None:
            condition = self.error_code.build_query_condition(
                contains_factory=SessionSchedulingHistoryConditions.by_error_code_contains,
                equals_factory=SessionSchedulingHistoryConditions.by_error_code_equals,
                starts_with_factory=SessionSchedulingHistoryConditions.by_error_code_starts_with,
                ends_with_factory=SessionSchedulingHistoryConditions.by_error_code_ends_with,
            )
            if condition:
                conditions.append(condition)

        if self.message is not None:
            condition = self.message.build_query_condition(
                contains_factory=SessionSchedulingHistoryConditions.by_message_contains,
                equals_factory=SessionSchedulingHistoryConditions.by_message_equals,
                starts_with_factory=SessionSchedulingHistoryConditions.by_message_starts_with,
                ends_with_factory=SessionSchedulingHistoryConditions.by_message_ends_with,
            )
            if condition:
                conditions.append(condition)

        if self.created_at is not None:
            condition = self.created_at.build_query_condition(
                before_factory=SessionSchedulingHistoryConditions.by_created_at_before,
                after_factory=SessionSchedulingHistoryConditions.by_created_at_after,
                equals_factory=SessionSchedulingHistoryConditions.by_created_at_equals,
            )
            if condition:
                conditions.append(condition)

        if self.updated_at is not None:
            condition = self.updated_at.build_query_condition(
                before_factory=SessionSchedulingHistoryConditions.by_updated_at_before,
                after_factory=SessionSchedulingHistoryConditions.by_updated_at_after,
                equals_factory=SessionSchedulingHistoryConditions.by_updated_at_equals,
            )
            if condition:
                conditions.append(condition)

        return conditions


@strawberry.input(description="Order by specification for session scheduling history")
class SessionSchedulingHistoryOrderBy(GQLOrderBy):
    field: SessionSchedulingHistoryOrderField
    direction: OrderDirection = OrderDirection.DESC

    @override
    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder."""
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case SessionSchedulingHistoryOrderField.CREATED_AT:
                return SessionSchedulingHistoryOrders.created_at(ascending)
            case SessionSchedulingHistoryOrderField.UPDATED_AT:
                return SessionSchedulingHistoryOrders.updated_at(ascending)


@strawberry.input(description="Filter for deployment history")
class DeploymentHistoryFilter(GQLFilter):
    id: UUIDFilter | None = None
    deployment_id: UUIDFilter | None = None
    phase: StringFilter | None = None
    from_status: list[str] | None = None
    to_status: list[str] | None = None
    result: list[SchedulingResultGQL] | None = None
    error_code: StringFilter | None = None
    message: StringFilter | None = None
    created_at: DateTimeFilter | None = None
    updated_at: DateTimeFilter | None = None

    @override
    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from this filter."""
        conditions: list[QueryCondition] = []

        if self.id is not None:
            condition = self.id.build_query_condition(
                equals_factory=DeploymentHistoryConditions.by_id_filter,
                in_factory=DeploymentHistoryConditions.by_id_in,
            )
            if condition:
                conditions.append(condition)

        if self.deployment_id is not None:
            condition = self.deployment_id.build_query_condition(
                equals_factory=DeploymentHistoryConditions.by_deployment_id_filter,
                in_factory=DeploymentHistoryConditions.by_deployment_id_in,
            )
            if condition:
                conditions.append(condition)

        if self.phase is not None:
            condition = self.phase.build_query_condition(
                contains_factory=DeploymentHistoryConditions.by_phase_contains,
                equals_factory=DeploymentHistoryConditions.by_phase_equals,
                starts_with_factory=DeploymentHistoryConditions.by_phase_starts_with,
                ends_with_factory=DeploymentHistoryConditions.by_phase_ends_with,
            )
            if condition:
                conditions.append(condition)

        if self.from_status is not None and len(self.from_status) > 0:
            conditions.append(DeploymentHistoryConditions.by_from_statuses(self.from_status))

        if self.to_status is not None and len(self.to_status) > 0:
            conditions.append(DeploymentHistoryConditions.by_to_statuses(self.to_status))

        if self.result is not None and len(self.result) > 0:
            conditions.append(
                DeploymentHistoryConditions.by_results([r.to_internal() for r in self.result])
            )

        if self.error_code is not None:
            condition = self.error_code.build_query_condition(
                contains_factory=DeploymentHistoryConditions.by_error_code_contains,
                equals_factory=DeploymentHistoryConditions.by_error_code_equals,
                starts_with_factory=DeploymentHistoryConditions.by_error_code_starts_with,
                ends_with_factory=DeploymentHistoryConditions.by_error_code_ends_with,
            )
            if condition:
                conditions.append(condition)

        if self.message is not None:
            condition = self.message.build_query_condition(
                contains_factory=DeploymentHistoryConditions.by_message_contains,
                equals_factory=DeploymentHistoryConditions.by_message_equals,
                starts_with_factory=DeploymentHistoryConditions.by_message_starts_with,
                ends_with_factory=DeploymentHistoryConditions.by_message_ends_with,
            )
            if condition:
                conditions.append(condition)

        if self.created_at is not None:
            condition = self.created_at.build_query_condition(
                before_factory=DeploymentHistoryConditions.by_created_at_before,
                after_factory=DeploymentHistoryConditions.by_created_at_after,
                equals_factory=DeploymentHistoryConditions.by_created_at_equals,
            )
            if condition:
                conditions.append(condition)

        if self.updated_at is not None:
            condition = self.updated_at.build_query_condition(
                before_factory=DeploymentHistoryConditions.by_updated_at_before,
                after_factory=DeploymentHistoryConditions.by_updated_at_after,
                equals_factory=DeploymentHistoryConditions.by_updated_at_equals,
            )
            if condition:
                conditions.append(condition)

        return conditions


@strawberry.input(description="Order by specification for deployment history")
class DeploymentHistoryOrderBy(GQLOrderBy):
    field: DeploymentHistoryOrderField
    direction: OrderDirection = OrderDirection.DESC

    @override
    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder."""
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case DeploymentHistoryOrderField.CREATED_AT:
                return DeploymentHistoryOrders.created_at(ascending)
            case DeploymentHistoryOrderField.UPDATED_AT:
                return DeploymentHistoryOrders.updated_at(ascending)


@strawberry.input(description="Filter for route history")
class RouteHistoryFilter(GQLFilter):
    id: UUIDFilter | None = None
    route_id: UUIDFilter | None = None
    deployment_id: UUIDFilter | None = None
    phase: StringFilter | None = None
    from_status: list[str] | None = None
    to_status: list[str] | None = None
    result: list[SchedulingResultGQL] | None = None
    error_code: StringFilter | None = None
    message: StringFilter | None = None
    created_at: DateTimeFilter | None = None
    updated_at: DateTimeFilter | None = None

    @override
    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from this filter."""
        conditions: list[QueryCondition] = []

        if self.id is not None:
            condition = self.id.build_query_condition(
                equals_factory=RouteHistoryConditions.by_id_filter,
                in_factory=RouteHistoryConditions.by_id_in,
            )
            if condition:
                conditions.append(condition)

        if self.route_id is not None:
            condition = self.route_id.build_query_condition(
                equals_factory=RouteHistoryConditions.by_route_id_filter,
                in_factory=RouteHistoryConditions.by_route_id_in,
            )
            if condition:
                conditions.append(condition)

        if self.deployment_id is not None:
            condition = self.deployment_id.build_query_condition(
                equals_factory=RouteHistoryConditions.by_deployment_id_filter,
                in_factory=RouteHistoryConditions.by_deployment_id_in,
            )
            if condition:
                conditions.append(condition)

        if self.phase is not None:
            condition = self.phase.build_query_condition(
                contains_factory=RouteHistoryConditions.by_phase_contains,
                equals_factory=RouteHistoryConditions.by_phase_equals,
                starts_with_factory=RouteHistoryConditions.by_phase_starts_with,
                ends_with_factory=RouteHistoryConditions.by_phase_ends_with,
            )
            if condition:
                conditions.append(condition)

        if self.from_status is not None and len(self.from_status) > 0:
            conditions.append(RouteHistoryConditions.by_from_statuses(self.from_status))

        if self.to_status is not None and len(self.to_status) > 0:
            conditions.append(RouteHistoryConditions.by_to_statuses(self.to_status))

        if self.result is not None and len(self.result) > 0:
            conditions.append(
                RouteHistoryConditions.by_results([r.to_internal() for r in self.result])
            )

        if self.error_code is not None:
            condition = self.error_code.build_query_condition(
                contains_factory=RouteHistoryConditions.by_error_code_contains,
                equals_factory=RouteHistoryConditions.by_error_code_equals,
                starts_with_factory=RouteHistoryConditions.by_error_code_starts_with,
                ends_with_factory=RouteHistoryConditions.by_error_code_ends_with,
            )
            if condition:
                conditions.append(condition)

        if self.message is not None:
            condition = self.message.build_query_condition(
                contains_factory=RouteHistoryConditions.by_message_contains,
                equals_factory=RouteHistoryConditions.by_message_equals,
                starts_with_factory=RouteHistoryConditions.by_message_starts_with,
                ends_with_factory=RouteHistoryConditions.by_message_ends_with,
            )
            if condition:
                conditions.append(condition)

        if self.created_at is not None:
            condition = self.created_at.build_query_condition(
                before_factory=RouteHistoryConditions.by_created_at_before,
                after_factory=RouteHistoryConditions.by_created_at_after,
                equals_factory=RouteHistoryConditions.by_created_at_equals,
            )
            if condition:
                conditions.append(condition)

        if self.updated_at is not None:
            condition = self.updated_at.build_query_condition(
                before_factory=RouteHistoryConditions.by_updated_at_before,
                after_factory=RouteHistoryConditions.by_updated_at_after,
                equals_factory=RouteHistoryConditions.by_updated_at_equals,
            )
            if condition:
                conditions.append(condition)

        return conditions


@strawberry.input(description="Order by specification for route history")
class RouteHistoryOrderBy(GQLOrderBy):
    field: RouteHistoryOrderField
    direction: OrderDirection = OrderDirection.DESC

    @override
    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder."""
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case RouteHistoryOrderField.CREATED_AT:
                return RouteHistoryOrders.created_at(ascending)
            case RouteHistoryOrderField.UPDATED_AT:
                return RouteHistoryOrders.updated_at(ascending)



from __future__ import annotations

import uuid
from collections.abc import Sequence
from functools import partial
from typing import Any, Self

import strawberry
from strawberry import ID
from strawberry.dataloader import DataLoader
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.common.data.artifact.types import ArtifactRegistryType
from ai.backend.manager.api.gql.utils import dedent_strip
from ai.backend.manager.data.artifact_registries.types import (
    ArtifactRegistryData,
)
from ai.backend.manager.errors.artifact_registry import ArtifactRegistryNotFoundError
from ai.backend.manager.services.artifact_registry.actions.common.get_multi import (
    GetArtifactRegistryMetasAction,
)

from .types import StrawberryGQLContext


@strawberry.type(
    description=dedent_strip("""
    Added in 25.15.0.

    Represents common metadata for an artifact registry.
    All artifact registry nodes expose that information regardless of type.
""")
)
class ArtifactRegistryMeta(Node):
    id: NodeID[str]
    name: str
    registry_id: ID
    type: ArtifactRegistryType
    url: str

    @classmethod
    def from_dataclass(cls, data: ArtifactRegistryData, url: str) -> Self:
        return cls(
            id=ID(str(data.id)),
            name=data.name,
            registry_id=ID(str(data.registry_id)),
            type=data.type,
            url=url,
        )

    @classmethod
    async def load_by_id(
        cls, ctx: StrawberryGQLContext, registry_ids: Sequence[uuid.UUID]
    ) -> list[Self]:
        from ai.backend.manager.api.gql.huggingface_registry import HuggingFaceRegistry
        from ai.backend.manager.api.gql.reservoir_registry import ReservoirRegistry

        # Get all registry metas in a single batch query
        registry_metas_action = (
            await ctx.processors.artifact_registry.get_registry_metas.wait_for_complete(
                GetArtifactRegistryMetasAction(registry_ids=list(registry_ids))
            )
        )

        # Create a mapping for efficient lookup
        registry_meta_map = {meta.registry_id: meta for meta in registry_metas_action.result}

        registries = []
        for registry_id in registry_ids:
            if registry_id not in registry_meta_map:
                raise ArtifactRegistryNotFoundError

            registry_meta = registry_meta_map[registry_id]
            registry_type = registry_meta.type
            url = None

            match registry_type:
                case ArtifactRegistryType.HUGGINGFACE:
                    hf_registry = await DataLoader(
                        partial(HuggingFaceRegistry.load_by_id, ctx),
                    ).load(registry_id)
                    url = hf_registry.url
                case ArtifactRegistryType.RESERVOIR:
                    reservoir_registry = await DataLoader(
                        partial(ReservoirRegistry.load_by_id, ctx),
                    ).load(registry_id)
                    url = reservoir_registry.endpoint

            registries.append(cls.from_dataclass(registry_meta, url=url))
        return registries


ArtifactRegistryMetaEdge = Edge[ArtifactRegistryMeta]


@strawberry.type(description="Added in 25.15.0")
class ArtifactRegistryMetaConnection(Connection[ArtifactRegistryMeta]):
    count: int

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count



"""Action for creating legacy deployments(Model Service)."""

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.deployment.creator import DeploymentCreationDraft
from ai.backend.manager.data.deployment.types import DeploymentInfo
from ai.backend.manager.services.deployment.actions.base import DeploymentBaseAction


@dataclass
class CreateLegacyDeploymentAction(DeploymentBaseAction):
    """Action to create a new legacy deployment(Model Service)."""

    draft: DeploymentCreationDraft

    @override
    def entity_id(self) -> str | None:
        return None  # New deployment doesn't have an ID yet

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class CreateLegacyDeploymentActionResult(BaseActionResult):
    data: DeploymentInfo

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id)



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.deployment.types import ModelDeploymentAutoScalingRuleData
from ai.backend.manager.repositories.base import BatchQuerier
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.base import (
    AutoScalingRuleBaseAction,
)


@dataclass
class SearchAutoScalingRulesAction(AutoScalingRuleBaseAction):
    querier: BatchQuerier

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class SearchAutoScalingRulesActionResult(BaseActionResult):
    data: list[ModelDeploymentAutoScalingRuleData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool

    @override
    def entity_id(self) -> str | None:
        return None



"""
ImageV2 GQL resolver for Strawberry GraphQL.

This module provides GraphQL query fields for ImageV2.
"""

from __future__ import annotations

from uuid import UUID

import strawberry
from strawberry import ID, Info

from ai.backend.common.types import ImageID
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.api.gql.utils import check_admin_only, dedent_strip
from ai.backend.manager.repositories.image.options import ImageAliasConditions, ImageConditions

from .fetcher import fetch_image, fetch_image_alias, fetch_image_aliases, fetch_images
from .types import (
    ContainerRegistryScopeGQL,
    ImageV2AliasConnectionGQL,
    ImageV2AliasFilterGQL,
    ImageV2AliasGQL,
    ImageV2AliasOrderByGQL,
    ImageV2ConnectionGQL,
    ImageV2FilterGQL,
    ImageV2GQL,
    ImageV2OrderByGQL,
    ImageV2ScopeGQL,
)


# Query Fields
@strawberry.field(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 26.2.0.

    Query images with optional filtering, ordering, and pagination (admin only).

    Returns container images available in the system. Images are container
    specifications that define runtime environments for compute sessions.

    Use filters to narrow down results by status, name, or architecture.
    Supports both cursor-based and offset-based pagination.
    """)
)
async def admin_images_v2(
    info: Info[StrawberryGQLContext],
    filter: ImageV2FilterGQL | None = None,
    order_by: list[ImageV2OrderByGQL] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> ImageV2ConnectionGQL | None:
    check_admin_only()
    return await fetch_images(
        info,
        filter,
        order_by,
        before,
        after,
        first,
        last,
        limit,
        offset,
    )


@strawberry.field(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 26.2.0.

    Retrieve a specific image by its ID.

    Returns detailed information about the image including its identity,
    metadata, resource requirements, and permission settings.
    """)
)
async def image_v2(id: ID, info: Info[StrawberryGQLContext]) -> ImageV2GQL | None:
    return await fetch_image(info, ImageID(UUID(id)))


@strawberry.field(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 26.2.0.

    Query images within a specific container registry with optional filtering,
    ordering, and pagination.

    Returns container images that belong to the specified registry.
    Use filters to narrow down results by status, name, or architecture.
    Supports both cursor-based and offset-based pagination.
    """)
)
async def container_registry_images_v2(
    info: Info[StrawberryGQLContext],
    scope: ContainerRegistryScopeGQL,
    filter: ImageV2FilterGQL | None = None,
    order_by: list[ImageV2OrderByGQL] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> ImageV2ConnectionGQL | None:
    # Add registry scope as base condition
    base_conditions = [ImageConditions.by_registry_id(scope.registry_id)]
    return await fetch_images(
        info,
        filter,
        order_by,
        before,
        after,
        first,
        last,
        limit,
        offset,
        base_conditions=base_conditions,
    )


# Image Alias Query Fields
@strawberry.field(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 26.2.0.

    Query image aliases with optional filtering, ordering, and pagination.

    Returns image aliases that provide alternative names for container images.
    Use filters to search by alias string.
    Supports both cursor-based and offset-based pagination.
    """)
)
async def admin_image_aliases(
    info: Info[StrawberryGQLContext],
    filter: ImageV2AliasFilterGQL | None = None,
    order_by: list[ImageV2AliasOrderByGQL] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> ImageV2AliasConnectionGQL | None:
    check_admin_only()
    return await fetch_image_aliases(
        info,
        filter,
        order_by,
        before,
        after,
        first,
        last,
        limit,
        offset,
    )


@strawberry.field(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 26.2.0.

    Retrieve a specific image alias by its ID.

    Returns the alias information including the alias string.
    """)
)
async def image_alias(id: ID, info: Info[StrawberryGQLContext]) -> ImageV2AliasGQL | None:
    return await fetch_image_alias(info, UUID(id))


@strawberry.field(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 26.2.0.

    Query image aliases within a specific image with optional filtering,
    ordering, and pagination.

    Returns image aliases that belong to the specified image.
    Supports both cursor-based and offset-based pagination.
    """)
)
async def image_scoped_aliases(
    info: Info[StrawberryGQLContext],
    scope: ImageV2ScopeGQL,
    filter: ImageV2AliasFilterGQL | None = None,
    order_by: list[ImageV2AliasOrderByGQL] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> ImageV2AliasConnectionGQL | None:
    # Add image scope as base condition
    base_conditions = [ImageAliasConditions.by_image_ids([ImageID(scope.image_id)])]
    return await fetch_image_aliases(
        info,
        filter,
        order_by,
        before,
        after,
        first,
        last,
        limit,
        offset,
        base_conditions=base_conditions,
    )



"""restore object permissions

Revision ID: ec7a778bcb78
Revises: 28fecac94e67
Create Date: 2025-08-07 20:34:59.442159

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "ec7a778bcb78"
down_revision = "28fecac94e67"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "object_permissions",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("status", sa.String(length=64), nullable=False, server_default="active"),
        sa.Column("role_id", GUID(), nullable=False),
        sa.Column("entity_type", sa.String(length=32), nullable=False),
        sa.Column("entity_id", sa.String(length=64), nullable=False),
        sa.Column("operation", sa.String(length=32), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_object_permissions")),
    )
    op.create_index(
        "ix_role_id_entity_id",
        "object_permissions",
        ["status", "role_id", "entity_id"],
        unique=False,
    )
    op.drop_constraint(
        "uq_scope_permissions_entity_operation_scope_id", "scope_permissions", type_="unique"
    )


def downgrade() -> None:
    op.create_unique_constraint(
        "uq_scope_permissions_entity_operation_scope_id",
        "scope_permissions",
        ["entity_type", "operation", "scope_id"],
    )
    op.drop_index("ix_role_id_entity_id", table_name="object_permissions")
    op.drop_table("object_permissions")



from __future__ import annotations

import logging
from collections.abc import Sequence
from datetime import datetime, timedelta
from uuid import UUID

from dateutil.relativedelta import relativedelta

from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.exception import (
    InvalidAPIParameters,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.models.resource_usage import (
    ProjectResourceUsage,
    parse_resource_usage_groups,
    parse_total_resource_group,
)
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.repositories.group.repositories import GroupRepositories
from ai.backend.manager.repositories.group.repository import GroupRepository
from ai.backend.manager.services.group.actions.create_group import (
    CreateGroupAction,
    CreateGroupActionResult,
)
from ai.backend.manager.services.group.actions.delete_group import (
    DeleteGroupAction,
    DeleteGroupActionResult,
)
from ai.backend.manager.services.group.actions.modify_group import (
    ModifyGroupAction,
    ModifyGroupActionResult,
)
from ai.backend.manager.services.group.actions.purge_group import (
    PurgeGroupAction,
    PurgeGroupActionResult,
)
from ai.backend.manager.services.group.actions.search_projects import (
    GetProjectAction,
    GetProjectActionResult,
    SearchProjectsAction,
    SearchProjectsActionResult,
    SearchProjectsByDomainAction,
    SearchProjectsByUserAction,
)
from ai.backend.manager.services.group.actions.usage_per_month import (
    UsagePerMonthAction,
    UsagePerMonthActionResult,
)
from ai.backend.manager.services.group.actions.usage_per_period import (
    UsagePerPeriodAction,
    UsagePerPeriodActionResult,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class GroupService:
    _config_provider: ManagerConfigProvider
    _valkey_stat_client: ValkeyStatClient
    _storage_manager: StorageSessionManager
    _group_repository: GroupRepository

    def __init__(
        self,
        storage_manager: StorageSessionManager,
        config_provider: ManagerConfigProvider,
        valkey_stat_client: ValkeyStatClient,
        group_repositories: GroupRepositories,
    ) -> None:
        self._storage_manager = storage_manager
        self._config_provider = config_provider
        self._valkey_stat_client = valkey_stat_client
        self._group_repository = group_repositories.repository

    async def create_group(self, action: CreateGroupAction) -> CreateGroupActionResult:
        group_data = await self._group_repository.create(action.creator)
        return CreateGroupActionResult(data=group_data)

    async def modify_group(self, action: ModifyGroupAction) -> ModifyGroupActionResult:
        # Convert user_uuids from list[str] to list[UUID] if provided
        user_uuids_converted = None
        user_uuids_list = action.user_uuids.optional_value()
        if user_uuids_list:
            user_uuids_converted = [UUID(user_uuid) for user_uuid in user_uuids_list]

        group_data = await self._group_repository.modify_validated(
            action.updater,
            action.user_update_mode.optional_value(),
            user_uuids_converted,
        )
        # If no group data is returned, it means only user updates were performed or no updates at all
        return ModifyGroupActionResult(data=group_data)

    async def delete_group(self, action: DeleteGroupAction) -> DeleteGroupActionResult:
        await self._group_repository.mark_inactive(action.group_id)
        return DeleteGroupActionResult(group_id=action.group_id)

    async def purge_group(self, action: PurgeGroupAction) -> PurgeGroupActionResult:
        await self._group_repository.purge_group(action.group_id)
        return PurgeGroupActionResult(group_id=action.group_id)

    async def _get_project_stats_for_period(
        self,
        start_date: datetime,
        end_date: datetime,
        project_ids: Sequence[UUID] | None = None,
    ) -> dict[UUID, ProjectResourceUsage]:
        kernels = await self._group_repository.fetch_project_resource_usage(
            start_date, end_date, project_ids=project_ids
        )
        local_tz = self._config_provider.config.system.timezone
        usage_groups = await parse_resource_usage_groups(
            kernels, self._valkey_stat_client, local_tz
        )
        total_groups, _ = parse_total_resource_group(usage_groups)
        return total_groups

    # group (or all the groups)
    async def usage_per_month(self, action: UsagePerMonthAction) -> UsagePerMonthActionResult:
        month = action.month
        local_tz = self._config_provider.config.system.timezone

        try:
            start_date = datetime.strptime(month, "%Y%m").replace(tzinfo=local_tz)
            end_date = start_date + relativedelta(months=+1)
        except ValueError as e:
            raise InvalidAPIParameters(extra_msg="Invalid date values") from e
        result = await self._group_repository.get_container_stats_for_period(
            start_date, end_date, action.group_ids
        )
        log.debug("container list are retrieved for month {0}", month)
        return UsagePerMonthActionResult(result=result)

    # group (or all the groups)
    async def usage_per_period(self, action: UsagePerPeriodAction) -> UsagePerPeriodActionResult:
        local_tz = self._config_provider.config.system.timezone
        project_id = action.project_id

        try:
            start_date = datetime.strptime(action.start_date, "%Y%m%d").replace(tzinfo=local_tz)
            end_date = datetime.strptime(action.end_date, "%Y%m%d").replace(tzinfo=local_tz)
            end_date = end_date + timedelta(days=1)  # include sessions in end_date
            if end_date - start_date > timedelta(days=100):
                raise InvalidAPIParameters("Cannot query more than 100 days")
        except ValueError as e:
            raise InvalidAPIParameters(extra_msg="Invalid date values") from e
        if end_date <= start_date:
            raise InvalidAPIParameters(extra_msg="end_date must be later than start_date.")
        log.info(
            "USAGE_PER_MONTH (p:{}, start_date:{}, end_date:{})", project_id, start_date, end_date
        )
        project_ids = [project_id] if project_id is not None else None
        usage_map = await self._get_project_stats_for_period(
            start_date, end_date, project_ids=project_ids
        )
        result = [p_usage.to_json(child=True) for p_usage in usage_map.values()]
        log.debug("container list are retrieved from {0} to {1}", start_date, end_date)
        return UsagePerPeriodActionResult(result=result)

    async def search_projects(self, action: SearchProjectsAction) -> SearchProjectsActionResult:
        """Search all projects (admin only - no scope filter).

        Args:
            action: SearchProjectsAction with querier.

        Returns:
            SearchProjectsActionResult with items and pagination info.
        """
        result = await self._group_repository.search_projects(querier=action.querier)
        return SearchProjectsActionResult(
            items=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def search_projects_by_domain(
        self, action: SearchProjectsByDomainAction
    ) -> SearchProjectsActionResult:
        """Search projects within a domain.

        Scope validation (domain existence) is done in repository layer.

        Args:
            action: SearchProjectsByDomainAction with scope and querier.

        Returns:
            SearchProjectsActionResult with domain-scoped items.
        """
        result = await self._group_repository.search_projects_by_domain(
            action.scope, action.querier
        )
        return SearchProjectsActionResult(
            items=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def search_projects_by_user(
        self, action: SearchProjectsByUserAction
    ) -> SearchProjectsActionResult:
        """Search projects a user is member of.

        Filters by association_groups_users table.

        Args:
            action: SearchProjectsByUserAction with scope and querier.

        Returns:
            SearchProjectsActionResult with user's projects.
        """
        result = await self._group_repository.search_projects_by_user(action.scope, action.querier)
        return SearchProjectsActionResult(
            items=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def get_project(self, action: GetProjectAction) -> GetProjectActionResult:
        """Get a single project by UUID.

        Args:
            action: GetProjectAction with project_id.

        Returns:
            GetProjectActionResult with project data.

        Raises:
            ProjectNotFound: If project does not exist.
        """
        data = await self._group_repository.get_project(action.project_id)
        return GetProjectActionResult(data=data)



from __future__ import annotations

import enum
import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Any


class ErrorLogSeverity(enum.StrEnum):
    CRITICAL = "critical"
    ERROR = "error"
    WARNING = "warning"


@dataclass
class ErrorLogMeta:
    created_at: datetime
    user: uuid.UUID | None
    source: str
    is_read: bool
    is_cleared: bool
    context_lang: str
    context_env: dict[str, Any]
    request_url: str | None
    request_status: int | None


@dataclass
class ErrorLogContent:
    severity: ErrorLogSeverity
    message: str
    traceback: str | None


@dataclass
class ErrorLogData:
    id: uuid.UUID
    meta: ErrorLogMeta
    content: ErrorLogContent


@dataclass
class ErrorLogListResult:
    """Search result with total count and pagination info for error logs."""

    items: list[ErrorLogData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool
